@@ -1,5 +1,5 @@
 /*	ani.h
-	Copyright (C) 2005-2006 Mark Tyler
+	Copyright (C) 2005-2009 Mark Tyler
 
 	This file is part of mtPaint.
 
@@ -18,23 +18,26 @@
 */
 
 #define MAX_CYC_SLOTS 100
-
+#define MAX_CYC_ITEMS 100
 #define MAX_FRAME 99999
-
 #define MAX_DELAY 1000
-
 #define ANI_PREFIX_LEN 16
 
-#define GIFSICLE_CREATE "gifsicle --colors 256 -w -O2 -D 2 -l0 --careful"
-	// global colourmaps, suppress warning, high optimizations, background removal method, infinite loops, ensure result works with Java & MS IE
+#if MAX_LAYERS > 128
+#error "Layer indices cannot fit in ani_cycle structure"
+#endif
+typedef struct {
+	int frame0, frame1, len;
+	signed char layers[MAX_CYC_ITEMS];
+} ani_cycle;
 
-
-int	ani_frame1, ani_frame2, ani_gif_delay;
+int ani_frame1, ani_frame2, ani_gif_delay;
+ani_cycle ani_cycle_table[MAX_CYC_SLOTS];
 
 
 void ani_init();			// Initialize variables/arrays etc.
 
-void pressed_animate_window( GtkMenuItem *menu_item, gpointer user_data );
+void pressed_animate_window();
 void pressed_set_key_frame();
 void pressed_remove_key_frames();
 
@@ -42,5 +45,3 @@ void ani_read_file( FILE *fp );		// Read data from layers file already opened
 void ani_write_file( FILE *fp );	// Write data to layers file already opened
 
 void ani_but_preview();			// Preview the animation
-
-int gifsicle( char *command );	// Execute Gifsicle/Gifview
