@@ -1,5 +1,5 @@
 /*	csel.c
-	Copyright (C) 2006-2008 Dmitry Groshev
+	Copyright (C) 2006-2011 Dmitry Groshev
 
 	This file is part of mtPaint.
 
@@ -17,10 +17,6 @@
 	along with mtPaint in the file COPYING.
 */
 
-#include <math.h>
-#include <stdlib.h>
-#include <gtk/gtk.h>
-
 #include "global.h"
 
 #include "mygtk.h"
@@ -28,6 +24,7 @@
 #include "otherwindow.h"
 #include "channels.h"
 #include "csel.h"
+#include "thread.h"
 
 /* Use sRGB gamma if defined, ITU-R 709 gamma otherwise */
 /* From my point of view, ITU-R 709 is a better model of a real CRT */
@@ -45,7 +42,7 @@ csel_info *csel_data;
 int csel_preview = 0x00FF00, csel_preview_a = 128, csel_overlay;
 
 double gamma256[256], gamma64[64];
-double midgamma256[256], midgamma64[64];
+double midgamma256[256];
 
 #ifndef NATIVE_DOUBLES
 float Fgamma256[256];
@@ -162,6 +159,21 @@ double rgb2B(double r, double g, double b)
 	return (y > 1.0 ? 1.0 : y);
 }
 
+#if 0 /* Disable while not in use */
+void rgb2Lab(double *tmp, double r, double g, double b)
+{
+	double x = r * rxyz[0] + g * gxyz[0] + b * bxyz[0];
+	double y = r * rxyz[1] + g * gxyz[1] + b * bxyz[1];
+	double z = r * rxyz[2] + g * gxyz[2] + b * bxyz[2];
+	double L = CIElum(y);
+	tmp[0] = L;
+	tmp[1] = (CIElum(x * (WHITE_Y / WHITE_X)) - L) * (500.0 / 116.0);
+	tmp[2] = (L - CIElum(z * (WHITE_Y / (1.0 - WHITE_X - WHITE_Y))) * (200.0 / 116.0);
+}
+#endif
+
+/* ITU-R 709 */
+
 #define RED_X 0.640
 #define RED_Y 0.330
 #define GREEN_X 0.300
@@ -169,10 +181,12 @@ double rgb2B(double r, double g, double b)
 #define BLUE_X 0.150
 #define BLUE_Y 0.060
 
-#if 1 // 6500K whitepoint
+#if 1 /* 6500K whitepoint (D65) */
 #define WHITE_X 0.3127
-#define WHITE_Y 0.329
-#else // 9300K whitepoint
+#define WHITE_Y 0.3290
+/* Some, like lcms, use y=0.3291, but that isn't how ITU-R BT.709-5 defines D65,
+ * but instead how SMPTE 240M does - WJ */
+#else /* 9300K whitepoint */
 #define WHITE_X 0.2848
 #define WHITE_Y 0.2932
 #endif
@@ -214,6 +228,7 @@ static void make_rgb_xyz(void)
 #define GAMMA_DIV 12.92321
 #endif
 #define KGAMMA 800
+#define KGAMMA64K 12900
 
 #else /* ITU-R 709 */
 
@@ -222,6 +237,7 @@ static void make_rgb_xyz(void)
 #define GAMMA_SPLIT 0.081
 #define GAMMA_DIV 4.5
 #define KGAMMA 300
+#define KGAMMA64K 4550
 
 #endif
 
@@ -244,8 +260,8 @@ static void make_gamma(double *Gamma, int cnt)
 	}
 }
 
-double kgamma256 = KGAMMA * 4, kgamma64 = KGAMMA;
-unsigned char ungamma256[KGAMMA * 4 + 1], ungamma64[KGAMMA + 1];
+int kgamma256 = KGAMMA * 4;
+unsigned char ungamma256[KGAMMA * 4 + 1];
 
 static void make_ungamma(double *Gamma, double *Midgamma,
 	unsigned char *Ungamma,	int kgamma, int cnt)
@@ -262,6 +278,63 @@ static void make_ungamma(double *Gamma, double *Midgamma,
 	for (; k <= kgamma; k++) Ungamma[k] = cnt - 1;
 }
 
+/* 16-bit gamma correction */
+
+static double gamma64K[255 * 4 + 2];
+static double gammaslope64K[255 * 4 + 1];
+static unsigned short ungamma64K[KGAMMA64K + 1];
+
+#if 0 /* Disable while not in use */
+double gamma65536(int idx)
+{
+	int n;
+
+	idx *= 4;
+	n = idx / 257;
+	return ((idx % 257) * (1.0 / 257.0) * (gamma64K[n + 1] - gamma64K[n]) +
+		gamma64K[n]);
+}
+
+int ungamma65536(double v)
+{
+	int n = ungamma64K[(int)(v * KGAMMA64K)];
+
+	n -= v < gamma64K[n];
+	return ((int)((n + (v - gamma64K[n]) * gammaslope64K[n]) * 64.25 + 0.5));
+}
+#endif
+
+double gamma65281(int idx)
+{
+	int n;
+
+	n = idx >> 6;
+	return ((idx & 0x3F) * (1.0 / 64.0) * (gamma64K[n + 1] - gamma64K[n]) +
+		gamma64K[n]);
+}
+
+int ungamma65281(double v)
+{
+	int n = ungamma64K[(int)(v * KGAMMA64K)];
+
+	n -= v < gamma64K[n];
+	return ((int)((n + (v - gamma64K[n]) * gammaslope64K[n]) * 64.0 + 0.5));
+}
+
+static void make_ungamma64K()
+{
+	int i, j, k = 0;
+
+	for (i = 1; i < 255 * 4 + 1; i++)
+	{
+		gammaslope64K[i - 1] = 1.0 / (gamma64K[i] - gamma64K[i - 1]);
+		j = gamma64K[i] * KGAMMA64K;
+		for (; k < j; k++) ungamma64K[k] = i - 1;
+	}
+	gammaslope64K[255 * 4] = 0.0;
+	for (; k <= KGAMMA64K; k++) ungamma64K[k] = 255 * 4;
+}
+
 static void make_CIE(void)
 {
 	int i;
@@ -285,10 +358,12 @@ static void make_EXP(void)
 
 void init_cols(void)
 {
+	make_gamma(gamma64K, 255 * 4 + 1);
+	gamma64K[255 * 4 + 1] = 1.0;
 	make_gamma(gamma256, 256);
 	make_gamma(gamma64, 64);
+	make_ungamma64K();
 	make_ungamma(gamma256, midgamma256, ungamma256, kgamma256, 256);
-	make_ungamma(gamma64, midgamma64, ungamma64, kgamma64, 64);
 	make_CIE();
 	make_EXP();
 	make_rgb_xyz();
@@ -333,7 +408,10 @@ int csel_scan(int start, int step, int cnt, unsigned char *mask,
 	unsigned char res = 0;
 	double d, dist = 0.0, lxn[3];
 	int i, j, k, l, jj, st3 = step * 3;
+	DEF_MUTEX(csel_lock); // To prevent concurrent writes to *info
 
+
+	LOCK_MUTEX(csel_lock);
 	cnt = start + step * (cnt - 1) + 1;
 	if (!mask)
 	{
@@ -459,6 +537,7 @@ int csel_scan(int start, int step, int cnt, unsigned char *mask,
 				mask[i] |= 255;
 		}
 	}
+	UNLOCK_MUTEX(csel_lock);
 	return (res);
 }
 
@@ -542,7 +621,7 @@ void csel_init()
 {
 	csel_info *info;
 
-	info = ALIGNTO(calloc(1, sizeof(csel_info) + sizeof(double)), double);
+	info = ALIGN(calloc(1, sizeof(csel_info) + sizeof(double)));
 	if (!info)
 	{
 		memory_errors(1);
