@@ -1,5 +1,5 @@
 /*	fpick.c
-	Copyright (C) 2007-2009 Mark Tyler and Dmitry Groshev
+	Copyright (C) 2007-2011 Mark Tyler and Dmitry Groshev
 
 	This file is part of mtPaint.
 
@@ -23,9 +23,6 @@
 
 #ifdef U_FPICK_MTPAINT		/* mtPaint fpicker */
 
-#include <dirent.h>
-#include <sys/stat.h>
-
 #include "inifile.h"
 #include "memory.h"
 #include "png.h"		// Needed by canvas.h
@@ -70,7 +67,6 @@ typedef struct
 	char		combo_items[FPICK_COMBO_ITEMS][PATHTXT],	// UTF8 in GTK+2
 			/* Must be DIR_SEP terminated at all times */
 			txt_directory[PATHBUF],	// Current directory - Normal C string
-			txt_file[PATHTXT],	// Full filename - Normal C string except on Windows
 			txt_mask[PATHTXT]	// Filter mask - UTF8 in GTK+2
 			;
 
@@ -105,144 +101,8 @@ static int case_insensitive;
 #define fpick_fnmatch(mask, str) !fnmatch(mask, str, FNM_PATHNAME | \
 	(case_insensitive ? FNM_CASEFOLD : 0))
 
-#elif (GTK_MAJOR_VERSION == 2) && (GTK_MINOR_VERSION < 4) /* GTK+ 2.0/2.2 */
-
-/*
- * fpick_fnmatch(), get_char() and get_unescaped_char() are based on code
- * from GTK+ 2.6.7
- * Copyright (C) 1991, 1992, 1993 Free Software Foundation, Inc.
- * Modified by the GTK+ Team and others 1997-2000.
- * Stripped down, converted to UTF-8 and test cases added by Owen Taylor, 2002.
- *
- * Adapted for mtPaint by Dmitry Groshev, November 2008.
- */
-
-#ifdef WIN32
-
-static gunichar get_char(const char **str)
-{
-	gunichar c = g_utf8_get_char(*str);
-	*str = g_utf8_next_char(*str);
-	return (g_unichar_tolower(c));
-}
-
-#define get_unescaped_char(str, was_escaped) get_char(str)
-
-#else
-
-static gunichar get_char(const char **str)
-{
-	gunichar c = g_utf8_get_char(*str);
-	*str = g_utf8_next_char(*str);
-	return (c);
-}
-
-static gunichar get_unescaped_char(const char **str, int *was_escaped)
-{
-	gunichar c = get_char (str);
-	if ((*was_escaped = c == '\\')) c = get_char(str);
-	return (c);
-}
-
-#endif /* !WIN32 */
-
-static int fpick_fnmatch(const char *pattern, const char *string)
-{
-	const char *p = pattern, *n = string;
-
-	while (*p)
-	{
-		const char *last_n = n;
-      		gunichar c = get_char(&p);
-		gunichar nc = get_char(&n);
-
-		switch (c)
-		{
-		case '?': if ((nc == '\0') || (nc == DIR_SEP)) return (FALSE);
-			break;
-		case '*':
-		{
-			const char *last_p;
-
-			while (TRUE)
-			{
-				last_p = p;
-				c = get_char(&p);
-				if (c == '?')
-				{
-					if ((nc == '\0') || (nc == DIR_SEP))
-						return (FALSE);
-					last_n = n;
-					nc = get_char(&n);
-				}
-				else if (c != '*') break;
-	      		}
-
-			/* If the pattern ends with wildcards, we have a
-			 * guaranteed match unless there is a dir separator
-			 * in the remainder of the string. */
-			if (c == '\0') return (!strchr(last_n, DIR_SEP));
-#ifndef WIN32
-			if (c == '\\') c = get_char(&p);
-#endif
-			for (p = last_p; nc != '\0';)
-			{
-/* !!! Can get exponential behaviour here, but I'm too lazy to fix this corner
- * case just right now - WJ */
-				if (((c == '[') || (c == nc)) &&
-					fpick_fnmatch(p, last_n)) return (TRUE);
-				last_n = n;
-				nc = get_char(&n);
-			}
-			return (FALSE);
-		}
-		case '[':
-		{
-			int not; // Inverted class flag
-			int was_escaped = FALSE;
-
-			if ((nc == '\0') || (nc == DIR_SEP)) return (FALSE);
-			if ((not = ((*p == '!') || (*p == '^')))) p++;
-
-			c = get_unescaped_char(&p, &was_escaped);
-			while (TRUE)
-			{
-				gunichar cstart = c, cend = c;
-
-				/* [ (unterminated) loses.  */
-				if (c == '\0') return (FALSE);
-				c = get_unescaped_char(&p, &was_escaped);
-				if (!was_escaped && (c == '-') && (*p != ']'))
-				{
-					cend = get_unescaped_char(&p, &was_escaped);
-					if (cend == '\0') return (FALSE);
-					c = get_unescaped_char(&p, &was_escaped);
-				}
-				if ((nc >= cstart) && (nc <= cend)) goto matched;
-				if (!was_escaped && (c == ']')) break;
-			}
-			if (!not) return (FALSE);
-			break;
-matched:		if (not) return (FALSE);
-			/* Skip the rest of the [...] that already matched.  */
-			while (was_escaped || (c != ']'))
-			{
-				/* [ (unterminated) loses.  */
-				if (c == '\0') return (FALSE);
-				c = get_unescaped_char(&p, &was_escaped);
-			}
-			break;
-	  	}
-#ifndef WIN32
-		case '\\': c = get_char(&p);
-		/* Fallthrough */
-#endif
-		default: if (c != nc) return (FALSE);
-		}
-	}
-	return (*n == '\0');
-}
-
+#elif (GTK_MAJOR_VERSION == 2) && (GTK2VERSION < 4) /* GTK+ 2.0/2.2 */
+#define fpick_fnmatch(mask, str) wjfnmatch(mask, str, TRUE)
 #endif
 
 /* !!! The code below manipulates undocumented internals of GtkCList, to work
@@ -275,7 +135,7 @@ static void fpick_clist_repattern(GtkCList *clist, const char *pattern)
 	GtkObject *obj = GTK_OBJECT(clist);
 	GList *tmp, *backup, *cur, *res, *pos = NULL;
 	int n = 0;
-#if (GTK_MAJOR_VERSION == 2) && (GTK_MINOR_VERSION >= 4) /* GTK+ 2.4+ */
+#if GTK2VERSION >= 4 /* GTK+ 2.4+ */
 	GtkFileFilter *filt = gtk_file_filter_new();
 	GtkFileFilterInfo info;
 	gtk_file_filter_add_pattern(filt, pattern);
@@ -310,7 +170,7 @@ static void fpick_clist_repattern(GtkCList *clist, const char *pattern)
 		tmp->prev = NULL;
 		/* Filter files, let directories pass */
 		if (GTK_CELL_TEXT(row->cell[FPICK_CLIST_SIZE])->text[0] && pattern[0] &&
-#if (GTK_MAJOR_VERSION == 2) && (GTK_MINOR_VERSION >= 4) /* GTK+ 2.4+ */
+#if GTK2VERSION >= 4 /* GTK+ 2.4+ */
 			(info.display_name = name , !gtk_file_filter_filter(filt, &info)))
 #else
 			!fpick_fnmatch(pattern, name))
@@ -333,7 +193,7 @@ static void fpick_clist_repattern(GtkCList *clist, const char *pattern)
 	clist_reselect_row(clist, n); // Avoid "select_row" signal emission
 	/* Let it be redrawn */
 	gtk_clist_thaw(clist);
-#if (GTK_MAJOR_VERSION == 2) && (GTK_MINOR_VERSION >= 4) /* GTK+ 2.4+ */
+#if GTK2VERSION >= 4 /* GTK+ 2.4+ */
 	gtk_object_sink(GTK_OBJECT(filt));
 #endif
 }
@@ -403,23 +263,21 @@ static char *isort_key(char *src)
 /* !!! Expects that "txt" points to PATHBUF-sized buffer */
 static void fpick_cleanse_path(char *txt)	// Clean up null terminated path
 {
-	static const char dds[] = { DIR_SEP, DIR_SEP, 0 };
 	char *src, *dest;
 
 #ifdef WIN32
 	// Unify path separators
-	src = txt;
-	while ((src = strchr(src, '/'))) *src = DIR_SEP;
+	reseparate(txt);
 #endif
 	// Expand home directory
 	if ((txt[0] == '~') && (txt[1] == DIR_SEP))
 	{
-		src = g_strconcat(get_home_directory(), txt + 1, NULL);
+		src = file_in_homedir(NULL, txt + 2, PATHBUF);
 		strncpy0(txt, src, PATHBUF - 1);
-		g_free(src);
+		free(src);
 	}
 	// Remove multiple consecutive occurences of DIR_SEP
-	if ((dest = src = strstr(txt, dds)))
+	if ((dest = src = strstr(txt, DIR_SEP_STR DIR_SEP_STR)))
 	{
 		while (*src)
 		{
@@ -595,12 +453,10 @@ static int fpick_scan_drives(fpicker *fp)	// Scan drives, populate widgets
 
 #endif
 
-static const char root_dir[] = { DIR_SEP, 0 };
-
-/* Scan directory, populate widgets */
+/* Scan directory, populate widgets; return 1 if success, 0 if total failure,
+ * -1 if failed with original dir and scanned a different one */
 static int fpick_scan_directory(fpicker *win, char *name, char *select)
 {
-	static char updir[] = { DIR_SEP, ' ', '.', '.', 0 };
 	static char nothing[] = "";
 	DIR	*dp;
 	struct	dirent *ep;
@@ -612,7 +468,7 @@ static int fpick_scan_directory(fpicker *win, char *name, char *select)
 	GtkCList  *clist = GTK_CLIST(win->clist);
 	GdkPixmap *icons[2];
 	GdkBitmap *masks[2];
-	int i, l, len, row, fail, idx = -1;
+	int i, l, len, row, fail, idx = -1, res = 1;
 
 
 	icons[0] = icons[1] = NULL;
@@ -640,6 +496,7 @@ static int fpick_scan_directory(fpicker *win, char *name, char *select)
 	fail = 0;
 	while (!(dp = opendir(full_name)))
 	{
+		res = -1; // Remember that original path was invalid
 		full_name[len - 1] = 0;
 		cp = strrchr(full_name, DIR_SEP);
 		// Try to go one level up
@@ -652,7 +509,7 @@ static int fpick_scan_directory(fpicker *win, char *name, char *select)
 			full_name[len++] = DIR_SEP;
                 }
 		// If current dir hasn't helped either, give up
-		else return (FALSE);
+		else return (0);
 		full_name[len] = 0;
 	}
 
@@ -676,9 +533,9 @@ static int fpick_scan_directory(fpicker *win, char *name, char *select)
 	gtk_clist_freeze(clist);
 	fpick_clist_clear(clist);	// Empty the list
 
-	if (strcmp(full_name, root_dir)) // Have a parent dir to move to?
+	if (strcmp(full_name, DIR_SEP_STR)) // Have a parent dir to move to?
 	{
-		row_txt[FPICK_CLIST_NAME] = updir;
+		row_txt[FPICK_CLIST_NAME] = DIR_SEP_STR " ..";
 		row_txt[FPICK_CLIST_TYPE] = row_txt[FPICK_CLIST_H_UC] =
 			row_txt[FPICK_CLIST_H_C] = "";
 		txt_size[0] = txt_date[0] = '\0';
@@ -734,7 +591,7 @@ static int fpick_scan_directory(fpicker *win, char *name, char *select)
 			while (src - tmp_txt >= 0) *dest-- = *src--;
 
 			cp = strrchr(txt_name, '.');
-			if (cp && (cp != txt_name))
+			if (cp && (cp != txt_name) && cp[1])
 			{
 #if GTK_MAJOR_VERSION == 1
 				g_strup(row_txt[FPICK_CLIST_TYPE] = g_strdup(cp + 1));
@@ -773,6 +630,8 @@ static int fpick_scan_directory(fpicker *win, char *name, char *select)
 	else fpick_clist_repattern(clist, win->txt_mask);
 	gtk_clist_thaw(clist);
 	fpick_clist_scroll(clist); // Scroll to selected row
+	// !!! Incomplete redraws only happen on Windows, but let's make sure
+	gtk_widget_queue_draw(win->clist);
 
 	closedir(dp);
 	gdk_pixmap_unref(icons[0]);
@@ -780,7 +639,7 @@ static int fpick_scan_directory(fpicker *win, char *name, char *select)
 	if (masks[0]) gdk_pixmap_unref(masks[0]);
 	if (masks[1]) gdk_pixmap_unref(masks[1]);
 
-	return TRUE;
+	return (res);
 }
 
 static void fpick_enter_dir_via_list(fpicker *fp, char *name)
@@ -795,15 +654,13 @@ static void fpick_enter_dir_via_list(fpicker *fp, char *name)
 		if (l && (ndir[l - 1] == DIR_SEP)) ndir[--l] = '\0';
 		c = strrchr(ndir, DIR_SEP);
 		if (c) *c = '\0';
-		else
-#ifndef WIN32
-			 strcpy(ndir, root_dir);
-#else
+		else /* Already in root directory */
 		{
+#ifdef WIN32
 			fpick_scan_drives(fp);
+#endif
 			return;
 		}
-#endif
 	}
 	else gtkncpy(ndir + l, name, PATHBUF - l);
 	fpick_cleanse_path(ndir);
@@ -846,31 +703,35 @@ static void fpick_select_row(GtkCList *clist, gint row, gint col,
 	}
 }
 
-static int fpick_enter_dirname(fpicker *fp, const char *name)
+/* Return 1 if changed directory, 0 if directory was the same, -1 if tried
+ * to change but failed */
+static int fpick_enter_dirname(fpicker *fp, const char *name, int l)
 {
 	char txt[PATHBUF], *ctxt;
+	int res = 0;
 
+	name = name ? g_strndup(name, l) :
+		g_strdup(gtk_entry_get_text(GTK_ENTRY(fp->combo_entry)));
 	gtkncpy(txt, name, PATHBUF);
 
 	fpick_cleanse_path(txt); // Path might have been entered manually
 
-	// Only do something if the directory is new
-	if (!strcmp(txt, fp->txt_directory)) return (0);
-
-	if (!fpick_scan_directory(fp, txt, NULL))
-	{	// Directory doesn't exist so ask user if they want to create it
+	if (strcmp(txt, fp->txt_directory) &&
+		// Only do something if the directory is new
+		((res = fpick_scan_directory(fp, txt, NULL)) <= 0))
+	{	// Directory doesn't exist so tell user
 		ctxt = g_strdup_printf(_("Could not access directory %s"), name);
-		alert_box(_("Error"), ctxt, _("OK"), NULL, NULL);
+		alert_box(_("Error"), ctxt, NULL);
 		g_free(ctxt);
-		return (-1);
+		res = res < 0 ? 1 : -1;
 	}
-	return (1);
+	g_free((char *)name);
+	return (res);
 }
 
 static void fpick_combo_changed(GtkWidget *widget, gpointer user_data)
 {
-	fpicker *fp = user_data;
-	fpick_enter_dirname(fp, gtk_entry_get_text(GTK_ENTRY(fp->combo_entry)));
+	fpick_enter_dirname(user_data, NULL, 0);
 }
 
 static void fpick_dialog_fn(char *key)
@@ -885,7 +746,7 @@ static void fpick_file_dialog(fpicker *fp, int row)
 	GtkWidget *win, *button, *label, *entry;
 	GtkCList *clist = GTK_CLIST(fp->clist);
 	GtkAccelGroup *ag = gtk_accel_group_new();
-	int l, res;
+	int uninit_(l), res;
 
 
 	if (row >= 0) /* Doing things to existing file */
@@ -984,7 +845,7 @@ static void fpick_file_dialog(fpicker *fp, int row)
 	}
 	g_free(snm);
 
-	if (tmp) alert_box(_("Error"), tmp, _("OK"), NULL, NULL);
+	if (tmp) alert_box(_("Error"), tmp, NULL);
 	else if (res > 1)
 	{
 		if (row >= 0) /* Deleted/renamed a file - move down */
@@ -1067,11 +928,11 @@ static gboolean fpick_click_event(GtkWidget *widget, GdkEventButton *event,
 #define _(X) X
 
 static toolbar_item fpick_bar[] = {
-	{ _("Up"),			-1, FPICK_ICON_UP, 0, xpm_up_xpm },
-	{ _("Home"),			-1, FPICK_ICON_HOME, 0, xpm_home_xpm },
-	{ _("Create New Directory"),	-1, FPICK_ICON_DIR, 0, xpm_newdir_xpm },
-	{ _("Show Hidden Files"),	 0, FPICK_ICON_HIDDEN, 0, xpm_hidden_xpm },
-	{ _("Case Insensitive Sort"),	 0, FPICK_ICON_CASE, 0, xpm_case_xpm },
+	{ _("Up"),			-1, FPICK_ICON_UP, 0, XPM_ICON(up) },
+	{ _("Home"),			-1, FPICK_ICON_HOME, 0, XPM_ICON(home) },
+	{ _("Create New Directory"),	-1, FPICK_ICON_DIR, 0, XPM_ICON(newdir) },
+	{ _("Show Hidden Files"),	 0, FPICK_ICON_HIDDEN, 0, XPM_ICON(hidden) },
+	{ _("Case Insensitive Sort"),	 0, FPICK_ICON_CASE, 0, XPM_ICON(case) },
 	{ NULL }};
 
 #undef _
@@ -1102,15 +963,20 @@ static void fpick_wildcard(GtkButton *button, gpointer user_data)
 	fpicker *fp = user_data;
 	char *ctxt, *ds, *nm, *mask = fp->txt_mask;
 
-	if (!fp->allow_files) return; // File entry is hidden anyway
 	ctxt = (char *)gtk_entry_get_text(GTK_ENTRY(fp->file_entry));
 	/* Presume filename if called by user pressing "OK", pattern otherwise */
 	if (button)
 	{
+		/* First, check if user had changed directory in the combo */
+		if (fpick_enter_dirname(fp, NULL, 0)) ctxt = NULL;
+		/* If file entry is hidden anyway */
+		else if (!fp->allow_files) return;
 		/* Filename must have some chars and no wildcards in it */
-		if (ctxt[0] && !ctxt[strcspn(ctxt, "?*")]) return;
+		else if (ctxt[0] && !ctxt[strcspn(ctxt, "?*")]) return;
 		/* Don't let pattern pass as filename */
 		gtk_signal_emit_stop_by_name(GTK_OBJECT(button), "clicked");
+		/* Don't do anything else for directory change */
+		if (!ctxt) return;
 	}
 
 	/* Do we have directory in here? */
@@ -1130,19 +996,16 @@ static void fpick_wildcard(GtkButton *button, gpointer user_data)
 		mask[l] = '\0';
 	}
 
-	while (ds) /* Have directory - enter it */
-	{
-		int res = fpick_enter_dirname(fp,
-			ctxt = g_strndup(ctxt, ds + 1 - ctxt));
-		g_free(ctxt);
-		if (res <= 0) break; // If failed to open a new dir
+	/* Have directory - enter it */
+	if (ds && (fpick_enter_dirname(fp, ctxt, ds + 1 - ctxt) > 0))
+	{	// Opened a new dir - skip redisplay
 		gtk_entry_set_text(GTK_ENTRY(fp->file_entry), nm); // Cut dir off
-		return; // Opened a new dir - skip redisplay
 	}
-
-	/* Torture GtkCList into displaying only files that match pattern */
-	fpick_clist_repattern(GTK_CLIST(fp->clist), mask);
-	fpick_clist_scroll(GTK_CLIST(fp->clist)); // Scroll to selected row
+	else
+	{	/* Torture GtkCList into displaying only files that match pattern */
+		fpick_clist_repattern(GTK_CLIST(fp->clist), mask);
+		fpick_clist_scroll(GTK_CLIST(fp->clist)); // Scroll to selected row
+	}
 }
 
 /* "Tab completion" for entry field, like in GtkFileSelection */
@@ -1181,15 +1044,14 @@ GtkWidget *fpick_create(char *title, int flags)		// Initialize file picker
 	res->sort_direction = GTK_SORT_ASCENDING;
 	res->sort_column = 0;
 	res->allow_files = res->allow_dirs = TRUE;
-	res->txt_directory[0] = res->txt_file[0] = '\0';
+	res->txt_directory[0] = '\0';
 
 	res->window = add_a_window( GTK_WINDOW_TOPLEVEL, title, GTK_WIN_POS_NONE, TRUE );
 	gtk_object_set_data(GTK_OBJECT(res->window), FP_DATA_KEY, res);
 
 	win_restore_pos(res->window, "fs_window", 0, 0, 550, 500);
 
-	vbox1 = gtk_vbox_new(FALSE, 0);
-	gtk_container_add(GTK_CONTAINER(res->window), vbox1);
+	vbox1 = add_vbox(res->window);
 	hbox1 = pack5(vbox1, gtk_hbox_new(FALSE, 0));
 
 	// ------- Combo Box -------
@@ -1358,42 +1220,19 @@ void fpick_set_filename(GtkWidget *fp, char *name, int raw)
 {
 	fpicker *win = gtk_object_get_data(GTK_OBJECT(fp), FP_DATA_KEY);
 	char txt[PATHTXT], *c;
-	int i = 0;
 
 	if (!raw)
 	{
 		/* Ensure that path is absolute */
-		txt[0] = '\0';
-		if ((name[0] != DIR_SEP)
-#ifdef WIN32
-			&& (name[0] != '/') && (name[1] != ':')
-#endif
-		)
-		{
-			getcwd(txt, PATHBUF - 1);
-			i = strlen(txt);
-			txt[i++] = DIR_SEP;
-		}
-
-		strnncat(txt, name, PATHTXT);
-#ifdef WIN32 /* Separators can be of both types on Windows - unify */
-		for (c = txt + i; (c = strchr(c, '/')); *c = DIR_SEP);
-#endif
-
+		resolve_path(txt, PATHBUF, name);
 		/* Separate the filename */
-		c = strrchr(txt, DIR_SEP);	// Guaranteed to be present now
-		name += c - txt - i + 1;
-		*c = '\0';
-#ifdef WIN32 /* Name is UTF8 on input there */
-		gtkncpy(txt + i, txt + i, PATHBUF - i);
-#endif
-
+		c = strrchr(txt, DIR_SEP);
+		*c++ = '\0';
 		// Scan directory, populate boxes if successful
 		if (!fpick_scan_directory(win, txt, "")) return;
 
-#ifndef WIN32 /* Name is in locale encoding on input */
-		name = gtkuncpy(txt, name, PATHTXT);
-#endif
+		/* Name is in locale encoding on input */
+		name = gtkuncpy(txt, c, PATHTXT);
 	}
 	gtk_entry_set_text(GTK_ENTRY(win->file_entry), name);
 }
@@ -1440,7 +1279,7 @@ static void fpick_iconbar_click(GtkWidget *widget, gpointer user_data)
 		break;
 	case FPICK_ICON_HOME:
 		gtkncpy(nm, gtk_entry_get_text(GTK_ENTRY(fp->file_entry)), PATHBUF);
-		snprintf(fnm, PATHBUF, "%s%c%s", get_home_directory(), DIR_SEP, nm);
+		file_in_homedir(fnm, nm, PATHBUF);
 		fpick_set_filename(fp->window, fnm, FALSE);
 		break;
 	case FPICK_ICON_DIR:
@@ -1467,30 +1306,26 @@ void fpick_setup(GtkWidget *fp, GtkWidget *xtra, GtkSignalFunc ok_fn,
 		"clicked", ok_fn, GTK_OBJECT(fp));
 	gtk_signal_connect_object(GTK_OBJECT(fpick->cancel_button),
 		"clicked", cancel_fn, GTK_OBJECT(fp));
-	gtk_signal_connect_object(GTK_OBJECT(fpick->window),
-		"delete_event", cancel_fn, GTK_OBJECT(fp));
+	delete_to_click(fpick->window, fpick->cancel_button);
 	pack(fpick->main_vbox, xtra);
 }
 
-const char *fpick_get_filename(GtkWidget *fp, int raw)
+void fpick_get_filename(GtkWidget *fp, char *buf, int len, int raw)
 {
 	fpicker *fpick = gtk_object_get_data(GTK_OBJECT(fp), FP_DATA_KEY);
 	char *txt = (char *)gtk_entry_get_text(GTK_ENTRY(fpick->file_entry));
-	char *dir = fpick->txt_directory;
 
-	if (raw) return (txt);
+	if (raw) strncpy0(buf, txt, len);
 #if GTK_MAJOR_VERSION == 1 /* Same encoding everywhere */
-	snprintf(fpick->txt_file, PATHTXT, "%s%s", dir, txt);
-#elif defined WIN32 /* Upconvert dir to UTF8 */
-	dir = gtkuncpy(NULL, dir, PATHTXT);
-	snprintf(fpick->txt_file, PATHTXT, "%s%s", dir, txt);
-	g_free(dir);
+	else snprintf(buf, len, "%s%s", fpick->txt_directory, txt);
 #else /* Convert filename to locale */
-	txt = gtkncpy(NULL, txt, PATHBUF);
-	snprintf(fpick->txt_file, PATHBUF, "%s%s", dir, txt);
-	g_free(txt);
+	else
+	{
+		txt = gtkncpy(NULL, txt, PATHBUF);
+		snprintf(buf, len, "%s%s", fpick->txt_directory, txt);
+		g_free(txt);
+	}
 #endif
-	return (fpick->txt_file);
 }
 #endif				/* mtPaint fpicker */
 
@@ -1531,8 +1366,7 @@ void fpick_setup(GtkWidget *fp, GtkWidget *xtra, GtkSignalFunc ok_fn,
 	gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(fp)->cancel_button),
 		"clicked", cancel_fn, GTK_OBJECT(fp));
 
-	gtk_signal_connect_object(GTK_OBJECT(fp),
-		"delete_event", cancel_fn, GTK_OBJECT(fp));
+	delete_to_click(fp, GTK_FILE_SELECTION(fp)->cancel_button);
 
 	pack(GTK_FILE_SELECTION(fp)->main_vbox, xtra);
 
@@ -1543,18 +1377,44 @@ void fpick_setup(GtkWidget *fp, GtkWidget *xtra, GtkSignalFunc ok_fn,
 #endif
 }
 
-const char *fpick_get_filename(GtkWidget *fp, int raw)
+void fpick_get_filename(GtkWidget *fp, char *buf, int len, int raw)
 {
-	if (raw) return (gtk_entry_get_text(GTK_ENTRY(
-		GTK_FILE_SELECTION(fp)->selection_entry)));
-	else return (gtk_file_selection_get_filename(GTK_FILE_SELECTION(fp)));
+	char *fname = (char *)gtk_entry_get_text(GTK_ENTRY(
+		GTK_FILE_SELECTION(fp)->selection_entry));
+	if (raw) strncpy0(buf, fname, len);
+	else
+	{
+#ifdef WIN32 /* Widget returns filename in UTF8 */
+		gtkncpy(buf, gtk_file_selection_get_filename(GTK_FILE_SELECTION(fp)), len);
+#else
+		strncpy0(buf, gtk_file_selection_get_filename(GTK_FILE_SELECTION(fp)), len);
+#endif
+		/* Make sure directory paths end with DIR_SEP */
+		if (fname[0]) return;
+		raw = strlen(buf);
+		if (!raw || (buf[raw - 1] != DIR_SEP))
+		{
+			if (raw > len - 2) raw = len - 2;
+			buf[raw] = DIR_SEP;
+			buf[raw + 1] = '\0';
+		}
+	}
 }
 
 void fpick_set_filename(GtkWidget *fp, char *name, int raw)
 {
 	if (raw) gtk_entry_set_text(GTK_ENTRY(
 		GTK_FILE_SELECTION(fp)->selection_entry), name);
+#ifdef WIN32 /* Widget wants filename in UTF8 */
+	else
+	{
+		name = gtkuncpy(NULL, name, 0);
+		gtk_file_selection_set_filename(GTK_FILE_SELECTION(fp), name);
+		g_free(name);
+	}
+#else
 	else gtk_file_selection_set_filename(GTK_FILE_SELECTION(fp), name);
+#endif
 }
 
 #endif		 /* GtkFileSelection based dialog */
