@@ -1,5 +1,5 @@
 /*	fpick.h
-	Copyright (C) 2007-2008 Mark Tyler and Dmitry Groshev
+	Copyright (C) 2007-2009 Mark Tyler and Dmitry Groshev
 
 	This file is part of mtPaint.
 
@@ -25,5 +25,5 @@ GtkWidget *fpick_create(char *title, int flags);
 void fpick_destroy(GtkWidget *fp);
 void fpick_setup(GtkWidget *fp, GtkWidget *xtra, GtkSignalFunc ok_fn,
 	GtkSignalFunc cancel_fn);
-const char *fpick_get_filename(GtkWidget *fp, int raw);
+void fpick_get_filename(GtkWidget *fp, char *buf, int len, int raw);
 void fpick_set_filename(GtkWidget *fp, char *name, int raw);
