@@ -1,5 +1,5 @@
 /*	info.c
-	Copyright (C) 2005-2008 Mark Tyler and Dmitry Groshev
+	Copyright (C) 2005-2011 Mark Tyler and Dmitry Groshev
 
 	This file is part of mtPaint.
 
@@ -31,25 +31,23 @@
 
 #define MAX_CUTS 128
 
-int	hs_rgb[256][3],			// Raw frequencies
-	hs_rgb_sorted[256][3],		// Sorted frequencies
-	hs_rgb_norm[256][3]		// Normalized frequencies
-	;
+static int hs_rgb[256][3],		// Raw frequencies
+	hs_rgb_sorted[256][3];		// Sorted frequencies
 
 #define HS_GRAPH_W 256
 #define HS_GRAPH_H 64
 
-unsigned char *hs_rgb_mem = NULL;	// RGB chunk holding graphs
+static unsigned char *hs_rgb_mem;	// RGB chunk holding graphs
 
-GtkWidget *hs_drawingarea;
-gboolean hs_norm;
+static GtkWidget *hs_drawingarea;
+static int hs_norm;
 
 
 static void hs_plot_graph()				// Plot RGB graphs
 {
 	unsigned char *im, col1[3] = { mem_pal_def[0].red, mem_pal_def[0].green, mem_pal_def[0].blue},
 			col2[3];
-	int i, j, k, t, min[3], max[3], med[3], bars[256][3];
+	int i, j, k, t, /*min[3],*/ max[3], med[3], bars[256][3];
 	float f;
 
 	for ( i=0; i<3; i++ ) col2[i] = 255 - col1[i];
@@ -103,7 +101,7 @@ static void hs_plot_graph()				// Plot RGB graphs
 				}
 			}
 
-			min[k] = hs_rgb_sorted[t][k];
+//			min[k] = hs_rgb_sorted[t][k];
 			med[k] = hs_rgb_sorted[(t+255)/2][k];
 			max[k] = hs_rgb_sorted[255][k];
 		}
@@ -213,12 +211,7 @@ static void hs_populate_rgb()				// Populate RGB tables
 		}
 	}
 
-	for ( i=0; i<256; i++ )
-	{
-		hs_rgb_sorted[i][0] = hs_rgb[i][0];
-		hs_rgb_sorted[i][1] = hs_rgb[i][1];
-		hs_rgb_sorted[i][2] = hs_rgb[i][2];
-	}
+	memcpy(hs_rgb_sorted, hs_rgb, sizeof(hs_rgb_sorted));
 
 	for ( k=0; k<3; k++ )			// Sort RGB table
 	{
@@ -237,7 +230,7 @@ static void hs_populate_rgb()				// Populate RGB tables
 	}
 }
 
-static gint hs_expose_graph( GtkWidget *widget, GdkEventExpose *event )
+static gboolean hs_expose_graph( GtkWidget *widget, GdkEventExpose *event )
 {
 	int x = event->area.x, y = event->area.y;
 	int w = event->area.width, h = event->area.height;
@@ -268,19 +261,13 @@ static gint hs_expose_graph( GtkWidget *widget, GdkEventExpose *event )
 
 ////	INFORMATION WINDOW
 
-GtkWidget *info_window;
+static GtkWidget *info_window;
 
-gint delete_info( GtkWidget *widget, GdkEvent *event, gpointer data )
+static void delete_info(GtkWidget *widget)
 {
-	gtk_widget_destroy(info_window);
-
-	if ( hs_rgb_mem != NULL )
-	{
-		free( hs_rgb_mem );
-		hs_rgb_mem = NULL;
-	}
-
-	return FALSE;
+	destroy_dialog(info_window);
+	free(hs_rgb_mem);
+	hs_rgb_mem = NULL;
 }
 
 void pressed_information()
@@ -288,22 +275,19 @@ void pressed_information()
 	char txt[256];
 	int i, j, orphans = 0, maxi;
 
-	GtkWidget *vbox4, *vbox5, *table4, *hs_normalize_check;
-	GtkWidget *scrolledwindow1, *viewport1, *table5, *button3;
-	GtkAccelGroup* ag = gtk_accel_group_new();
+	GtkWidget *vbox4, *vbox5, *table4;//, *hs_normalize_check;
+	GtkWidget *scrolledwindow1, *viewport1, *table5;
 
 	info_window = add_a_window( GTK_WINDOW_TOPLEVEL, _("Information"), GTK_WIN_POS_CENTER, TRUE );
 
 	if ( mem_img_bpp == 1 )
 		gtk_widget_set_usize (GTK_WIDGET (info_window), -2, 400);
 
-	vbox4 = gtk_vbox_new (FALSE, 0);
-	gtk_widget_show (vbox4);
-	gtk_container_add (GTK_CONTAINER (info_window), vbox4);
+	vbox4 = add_vbox(info_window);
 
 	table4 = gtk_table_new (3, 2, FALSE);
 	gtk_widget_show (table4);
-	add_with_frame(vbox4, _("Memory"), table4, 5);
+	add_with_frame(vbox4, _("Memory"), table4);
 	gtk_container_set_border_width (GTK_CONTAINER (table4), 5);
 
 	add_to_table( _("Total memory for main + undo images"), table4, 0, 0, 5 );
@@ -363,7 +347,7 @@ void pressed_information()
 
 	vbox5 = gtk_vbox_new (FALSE, 0);
 	gtk_widget_show (vbox5);
-	add_with_frame(vbox4, _("Colour Histogram"), vbox5, 4);
+	add_with_frame_x(vbox4, _("Colour Histogram"), vbox5, 4, FALSE);
 
 	hs_norm = FALSE;
 	hs_rgb_mem = malloc( HS_GRAPH_W * HS_GRAPH_H * 3 * mem_img_bpp );
@@ -376,7 +360,8 @@ void pressed_information()
 	gtk_signal_connect_object( GTK_OBJECT(hs_drawingarea), "expose_event",
 		GTK_SIGNAL_FUNC (hs_expose_graph), NULL );
 
-	hs_normalize_check = pack(vbox5, sig_toggle(_("Normalize"), FALSE, NULL,
+//	hs_normalize_check =
+	pack(vbox5, sig_toggle(_("Normalize"), FALSE, NULL,
 		GTK_SIGNAL_FUNC(hs_click_normalize)));
 
 	if ( mem_img_bpp == 1 )
@@ -432,17 +417,9 @@ void pressed_information()
 	}
 
 	add_hseparator( vbox4, -2, 10 );
+	pack(vbox4, OK_box(0, info_window, _("OK"), GTK_SIGNAL_FUNC(delete_info),
+		NULL, NULL));
 
-	button3 = add_a_button(_("OK"), 2, vbox4, FALSE);
-	gtk_widget_add_accelerator (button3, "clicked", ag, GDK_Escape, 0, (GtkAccelFlags) 0);
-	gtk_widget_add_accelerator (button3, "clicked", ag, GDK_Return, 0, (GtkAccelFlags) 0);
-	gtk_widget_add_accelerator (button3, "clicked", ag, GDK_KP_Enter, 0, (GtkAccelFlags) 0);
-	gtk_signal_connect(GTK_OBJECT(button3), "clicked",
-		GTK_SIGNAL_FUNC (delete_info), GTK_OBJECT (info_window) );
-	gtk_signal_connect_object (GTK_OBJECT (info_window), "delete_event",
-		GTK_SIGNAL_FUNC (delete_info), GTK_OBJECT (info_window));
-
-	gtk_widget_show (info_window);
-	gtk_window_set_transient_for( GTK_WINDOW(info_window), GTK_WINDOW(main_window) );
-	gtk_window_add_accel_group(GTK_WINDOW (info_window), ag);
+	gtk_window_set_transient_for(GTK_WINDOW(info_window), GTK_WINDOW(main_window));
+	gtk_widget_show(info_window);
 }
