@@ -1,5 +1,5 @@
 /*	inifile.c
-	Copyright (C) 2007-2009 Dmitry Groshev
+	Copyright (C) 2007-2011 Dmitry Groshev
 
 	This file is part of mtPaint.
 
@@ -34,6 +34,7 @@
 #include <gtk/gtk.h>
 
 #include "global.h"
+#include "memory.h"
 #include "inifile.h"
 
 /* Make code not compile where it cannot run */
@@ -67,6 +68,83 @@ typedef char Integers_Do_Not_Fit_Into_Pointers[2 * (sizeof(int) <= sizeof(char *
 
 #define SLOT_NAME(I,S) ((I)->sblock[(S)->flags & INI_STRING] + (S)->key)
 
+/* Escaping is needed for leading tabs and spaces, and for inline CR and LF */
+static char *escape_string(char *src)
+{
+	static const char *escaped = "\t\r\n \\";
+	char c, *cp, *tmp, *t2;
+
+	if ((src[0] != '\t') && (src[0] != ' ') && !src[strcspn(src, "\r\n")])
+		return (NULL);
+	t2 = tmp = calloc(1, strlen(src) * 2 + 1);
+	if (tmp)
+	{
+		while ((c = *src++))
+		{
+			if ((cp = strchr(escaped, c)))
+			{
+				*t2++ = '\\';
+				c = "trn \\"[cp - escaped];
+			}
+			*t2++ = c;
+		}
+		*t2 = '\0';
+	}
+	return (tmp);
+}
+
+#if GTK_MAJOR_VERSION == 1
+
+/* GLib 1.2 doesn't provide g_strcompress() */
+static void unescape_string(char *buf)
+{
+#define NUM_ESCAPES 5
+	static const char escapes[] = "bfnrt01234567";
+	static const char escaped[] = { 7, 12, 10, 13, 8, 0, 1, 2, 3, 4, 5, 6, 7 };
+	char c, cc, *tmp, *src = buf;
+	int v;
+
+	while ((c = *src++))
+	{
+		if (c == '\\')
+		{
+			c = *src++;
+			if (!c) break;
+			while ((tmp = strchr(escapes, c)))
+			{
+				v = tmp - escapes;
+				c = escaped[v];
+				if (v >= NUM_ESCAPES)
+				{
+					// Octal escape
+					cc = *src - '0';
+					if (cc & ~7) break;
+					c = c * 8 + cc;
+					cc = *(++src) - '0';
+					if (cc & ~7) break;
+					c = c * 8 + cc;
+					++src;
+				}
+				break;
+			}
+		}
+		*buf++ = c;
+	}
+	*buf = '\0';
+#undef NUM_ESCAPES
+}
+
+#else
+
+static void unescape_string(char *buf)
+{
+	char *tmp = g_strcompress(buf);
+	strcpy(buf, tmp);
+	g_free(tmp);
+}
+
+#endif
+
 /* Thomas Wang's and "One at a time" hash functions */
 static guint32 hashf(guint32 seed, int section, char *key)
 {
@@ -125,13 +203,7 @@ static int resize_hash(inifile *inip, int cnt)
 	int len;
 
 	if (!cnt) return (0); /* Degenerate case */
-	len = HASHFILL(cnt) - 1;
-	len |= len >> 1;
-	len |= len >> 2;
-	len |= len >> 4;
-	len |= len >> 8;
-	len |= len >> 16;
-	len++;
+	len = nextpow2(HASHFILL(cnt) - 1);
 	if (len <= inip->hmask * 2 + 2) return (0); /* Large enough */
 	free(inip->hash);
 	inip->hash = calloc(1, len * sizeof(gint32));
@@ -328,7 +400,7 @@ int read_ini(inifile *inip, char *fname, int itype)
 	inifile ini;
 	inislot *slot;
 	char *tmp, *wrk, *w2, *str;
-	int i, j, l, sec = 0;
+	int i, j, l, q, sec = 0;
 
 
 	/* Read the file */
@@ -383,7 +455,8 @@ error:			g_printerr("Wrong INI line: '%s'\n", tmp);
 		for (wrk = w2 - 1; wrk - tmp >= 0; wrk--)
 			if ((*wrk != ' ') && (*wrk != '\t')) break;
 		wrk[1] = '\0';
-		w2 += 1 + strspn(w2 + 1, "\t ");
+		q = *(++w2) == '='; /* "==" means quoted value */
+		w2 += q + strspn(w2 + q, "\t ");
 //		for (wrk = str - 1; *wrk && ((*wrk == '\t') || (*wrk == ' '); *wrk-- = '\0');
 		/* Hash this pair */
 		slot = cuckoo_find(&ini, sec, tmp);
@@ -391,13 +464,14 @@ error:			g_printerr("Wrong INI line: '%s'\n", tmp);
 		{
 			slot = add_slot(&ini);
 			if (!slot) goto fail;
-			slot->type = INI_UNDEF;
 			slot->sec = sec;
 			slot->key = tmp - ini.sblock[itype];
 			slot->flags = itype;
 			if (!cuckoo_insert(&ini, slot) && !rehash(&ini))
 				goto fail;
 		}
+		if (q) unescape_string(w2);
+		slot->type = INI_UNDEF;
 		slot->value = w2;
 	}
 
@@ -415,7 +489,7 @@ int write_ini(inifile *inip, char *fname, char *header)
 {
 	FILE *fp;
 	inislot *slotp;
-	char *name, *sv;
+	char *name, *sv, *xv;
 	int i, j, max, sec, var, defv;
 
 	if (!(fp = fopen(fname, "w"))) return (FALSE);
@@ -457,7 +531,13 @@ int write_ini(inifile *inip, char *fname, char *header)
 					slotp->defv, sv)) break;
 			case INI_UNDEF:
 			default:
-				fprintf(fp, "%s = %s\n", name, sv);
+				/* Escape value if needed */
+				if ((xv = escape_string(sv)))
+				{
+					fprintf(fp, "%s == %s\n", name, xv);
+					free(xv);
+				}
+				else fprintf(fp, "%s = %s\n", name, sv);
 				break;
 			case INI_INT:
 				if (defv && ((int)slotp->value == slotp->defv)) break;
@@ -695,6 +775,22 @@ char *get_home_directory(void)
 	return homedir;
 }
 
+char *extend_path(const char *path)
+{
+	char *dir, *name;
+
+	if (path[0] == '~')
+		return (g_strdup_printf("%s%s", get_home_directory(), path + 1));
+	if (path[0] && (path[1] == ':')) // Path w/drive letter is absolute
+		return (g_strdup(path));
+	name = g_win32_get_package_installation_directory(NULL, NULL);
+	dir = g_locale_from_utf8(name, -1, NULL, NULL, NULL);
+	g_free(name);
+	name = g_strdup_printf("%s%s", dir, path);
+	g_free(dir);
+	return (name);
+}
+
 #else
 
 #include <unistd.h>
@@ -725,6 +821,13 @@ gchar *get_home_directory(void)
 	return homedir;
 }
 
+char *extend_path(const char *path)
+{
+	if (path[0] == '~')
+		return (g_strdup_printf("%s%s", get_home_directory(), path + 1));
+	return (g_strdup(path));
+}
+
 #endif
 
 /* Compatibility functions */
@@ -732,26 +835,33 @@ gchar *get_home_directory(void)
 static inifile main_ini;
 static char *main_ininame;
 
-void inifile_init(char *ini_filename)
+void inifile_init(char *system_ini, char *user_ini)
 {
+	char *tmp, *ini = user_ini;
 	int res, mask = 3;
 
-	main_ininame = g_strdup_printf("%s%s", get_home_directory(), ini_filename);
 	while (new_ini(&main_ini))
 	{
-#ifndef WIN32
-		if (mask & 1)
+		if ((mask & 1) && system_ini)
 		{
-			res = read_ini(&main_ini, "/etc/mtpaint/mtpaintrc", INI_SYSTEM);
+			tmp = extend_path(system_ini);
+			res = read_ini(&main_ini, tmp, INI_SYSTEM);
+			g_free(tmp);
 			if (res <= 0) mask ^= 1; // Don't try again if failed
 			if (res < 0) continue; // Restart if struct got deleted
+			/* !!! Allow system inifile to relocate user inifile */
+			ini = inifile_get("userINI", user_ini);
+			if (!ini[0]) ini = system_ini;
 		}
-#endif
-		if (mask & 2)
+		if ((mask & 2) && user_ini)
 		{
+			main_ininame = extend_path(ini);
 			res = read_ini(&main_ini, main_ininame, INI_USER);
-			if (res <= 0) mask ^= 2;
-			if (res < 0) continue;
+			if (res <= 0) // Failed
+			{
+				mask ^= 2;
+				if (res < 0) continue;
+			}
 		}
 		break;
 	}
