@@ -1,5 +1,5 @@
 /*	inifile.h
-	Copyright (C) 2007-2009 Dmitry Groshev
+	Copyright (C) 2007-2011 Dmitry Groshev
 
 	This file is part of mtPaint.
 
@@ -53,10 +53,11 @@ int ini_getsection(inifile *inip, int section, char *key);
 
 char *slurp_file(char *fname);
 char *get_home_directory(void);
+char *extend_path(const char *path);
 
 /* Compatibility functions */
 
-void inifile_init(char *ini_filename);
+void inifile_init(char *system_ini, char *user_ini);
 void inifile_quit();
 
 char *inifile_get(char *setting, char *defaultValue);
