@@ -1,5 +1,5 @@
 /*	layer.c
-	Copyright (C) 2005-2009 Mark Tyler and Dmitry Groshev
+	Copyright (C) 2005-2011 Mark Tyler and Dmitry Groshev
 
 	This file is part of mtPaint.
 
@@ -73,15 +73,14 @@ void layers_init()
 
 /* Allocate layer image, its channels and undo stack
  * !!! Must be followed by update_undo() after setting up image is done */
-layer_image *alloc_layer(int w, int h, int bpp, int cmask, chanlist src)
+layer_image *alloc_layer(int w, int h, int bpp, int cmask, image_info *src)
 {
 	layer_image *lim;
 
-	if (src) cmask = cmask_from(src);
 	lim = calloc(1, sizeof(layer_image));
 	if (!lim) return (NULL);
 	if (init_undo(&lim->image_.undo_, mem_undo_depth) &&
-		mem_alloc_image(AI_COPY, &lim->image_, w, h, bpp, cmask, src))
+		mem_alloc_image(src ? AI_COPY : 0, &lim->image_, w, h, bpp, cmask, src))
 		return (lim);
 	free(lim->image_.undo_.items);
 	free(lim);
@@ -118,14 +117,15 @@ static void repaint_layers()
 
 ///	LAYERS WINDOW
 
-GtkWidget *layers_window = NULL;
+GtkWidget *layers_box, *layers_window;
 
 typedef struct {
 	GtkWidget *item, *name, *toggle;
 } layer_item;
 
-static GtkWidget *layer_list, *entry_layer_name, *layer_x, *layer_y,
-	*layer_tools[TOTAL_ICONS_LAYER], *layer_spin, *layer_slider;
+static GtkWidget *layer_list, *entry_layer_name,
+	*layer_x, *layer_y, *layer_spin, *layer_slider,
+	*layer_tools[TOTAL_ICONS_LAYER];
 static layer_item layer_list_data[MAX_LAYERS + 1];
 
 static int layers_initialized;		// Indicates if initializing is complete
@@ -135,26 +135,27 @@ static int layers_sensitive(int state)
 {
 	int res;
 
-	if (!layers_window) return (TRUE);
-	res = GTK_WIDGET_SENSITIVE(layers_window);
-	gtk_widget_set_sensitive(layers_window, state);
+	if (!layers_box) return (TRUE);
+	res = GTK_WIDGET_SENSITIVE(layers_box);
+	gtk_widget_set_sensitive(layers_box, state);
+	/* Delayed action - selection could not be moved while insensitive */
+	if (!res && state)
+		list_select_item(layer_list, layer_list_data[layer_selected].item);
 	return (res);
 }
 
 static void layers_update_titlebar()		// Update filename in titlebar
 {
-	char txt[300], txt2[PATHTXT], *extra = "-";
+	char txt[300], txt2[PATHTXT];
 
-	if ( layers_window == NULL ) return;		// Don't bother if window is not showing
+
+	if (!layers_window) return;	// Don't bother if window is not showing
 
 	gtkuncpy(txt2, layers_filename, PATHTXT);
-
-	if ( layers_changed == 1 ) extra = _("(Modified)");
-
-	snprintf( txt, 290, "%s %s %s", _("Layers"), extra, txt2[0] ? txt2 :
-		_("Untitled"));
-
-	gtk_window_set_title (GTK_WINDOW (layers_window), txt );
+	snprintf(txt, 290, "%s %s %s", _("Layers"),
+		layers_changed ? _("(Modified)") : "-",
+		txt2[0] ? txt2 : _("Untitled"));
+	gtk_window_set_title(GTK_WINDOW(layers_window), txt);
 }
 
 void layers_notify_changed()			// Layers have just changed - update vars as needed
@@ -194,26 +195,11 @@ void layer_copy_to_main( int l )		// Copy info from layer to main image
 	mem_state = lp->state_;
 }
 
-/* !!! An evil hack for reliably moving focus around in GtkList */
 static void layer_select_slot(int slot)
 {
-	GtkWidget *item, *fw;
-
-	if (!layers_window) return;
-
-	item = layer_list_data[slot].item;
-	fw = GTK_WINDOW(layers_window)->focus_widget;
-
-	/* Focus is somewhere in list - move it, selection will follow */
-	if (fw && gtk_widget_is_ancestor(fw, layer_list))
-		gtk_widget_grab_focus(item);
-	else /* Focus is elsewhere - move whatever remains, then */
-	{
-	/* !!! For simplicity, an undocumented field is used; a bit less hacky
-	 * but longer is to set focus child to item, NULL, and item again - WJ */
-		gtk_container_set_focus_child(GTK_CONTAINER(layer_list), item);
-		GTK_LIST(layer_list)->last_focus_child = item;
-	}
+	/* !!! Selection gets stuck if tried to move it while insensitive */
+	if (layers_box && GTK_WIDGET_SENSITIVE(layers_box))
+		list_select_item(layer_list, layer_list_data[slot].item);
 }
 
 void shift_layer(int val)
@@ -267,7 +253,7 @@ void layer_show_new()
 	update_main_with_new_layer();
 	layers_notify_changed();
 
-	if (layers_window)
+	if (layers_box)
 	{
 		layer_item *l = layer_list_data + layers_total;
 		layer_node *t = layer_table + layers_total;
@@ -301,12 +287,13 @@ int layer_add(int w, int h, int bpp, int cols, png_color *pal, int cmask)
 		memory_errors(1);
 		return (FALSE);
 	}
-	lim->state_.xpm_trans = lim->state_.xbm_hot_x = lim->state_.xbm_hot_y = -1;
+	lim->state_.xbm_hot_x = lim->state_.xbm_hot_y = -1;
 	lim->state_.channel = lim->image_.img[mem_channel] ? mem_channel : CHN_IMAGE;
 
+	lim->image_.trans = -1;
 	lim->image_.cols = cols;
 	if (pal) mem_pal_copy(lim->image_.pal, pal);
-	else mem_scale_pal(lim->image_.pal, 0, 0,0,0, cols - 1, 255,255,255);
+	else mem_bw_pal(lim->image_.pal, 0, cols - 1);
 
 	init_istate(&lim->state_, &lim->image_);
 	update_undo(&lim->image_);
@@ -335,7 +322,7 @@ void layer_press_duplicate()
 
 	if (layers_total >= MAX_LAYERS) return;
 
-	lim = alloc_layer(mem_width, mem_height, mem_img_bpp, 0, mem_img);
+	lim = alloc_layer(0, 0, 0, 0, &mem_image);
 	if (!lim)
 	{
 		memory_errors(1);
@@ -352,6 +339,7 @@ void layer_press_duplicate()
 	lim->state_ = ls->state_;
 	mem_pal_copy(lim->image_.pal, ls->image_.pal);
 	lim->image_.cols = ls->image_.cols;
+	lim->image_.trans = ls->image_.trans;
 	update_undo(&lim->image_);
 
 	// Copy across position data
@@ -380,7 +368,7 @@ void layer_refresh_list()
 	static int in_refresh;
 	int i;
 
-	if (!layers_window) return;
+	if (!layers_box) return;
 
 	// As befits a refresh, this should show the final state
 	if (in_refresh)
@@ -431,7 +419,7 @@ void layer_press_delete()
 	snprintf(txt, 256, _("Do you really want to delete layer %i (%s) ?"),
 		layer_selected, layer_table[layer_selected].name );
 
-	i = alert_box( _("Warning"), txt, _("No"), _("Yes"), NULL );
+	i = alert_box(_("Warning"), txt, _("No"), _("Yes"), NULL);
 	if ((i != 2) || (check_for_changes() == 1)) return;
 
 	layer_copy_from_main(layer_selected);
@@ -449,7 +437,7 @@ static void layer_show_position()
 	layer_node *t = layer_table + layer_selected;
 	int oldinit = layers_initialized;
 
-	if (!layers_window) return;
+	if (!layers_box) return;
 
 	layers_initialized = FALSE;
 	gtk_spin_button_set_value(GTK_SPIN_BUTTON(layer_x), t->x);
@@ -457,6 +445,23 @@ static void layer_show_position()
 	layers_initialized = oldinit;
 }
 
+void layer_show_trans()
+{
+	GtkSpinButton *spin;
+	int n;
+
+	if (!layers_box) return;
+	spin = GTK_SPIN_BUTTON(layer_spin);
+	n = layer_selected ? mem_xpm_trans : -1;
+	if (gtk_spin_button_get_value_as_int(spin) != n)
+	{
+		int oldinit = layers_initialized;
+		layers_initialized = FALSE;
+		gtk_spin_button_set_value(spin, n);
+		layers_initialized = oldinit;
+	}
+}
+
 void layer_press_centre()
 {
 	if (!layer_selected) return; // Nothing to do
@@ -469,46 +474,29 @@ void layer_press_centre()
 	repaint_layers();
 }
 
-static int layers_unsaved_tot()	// Return number of layers with no filenames
+/* Return 1 if some layers are modified, 2 if some are nameless, 3 if both,
+ * 0 if neither */
+static int layers_changed_tot()
 {
-	int j = 0, k;
-
-	for ( k=0; k<=layers_total; k++ )	// Check each layer for proper filename
-	{
-		j += !(k == layer_selected ? mem_filename[0] :
-			layer_table[k].image->state_.filename[0]);
-	}
-
-	return j;
-}
-
-static int layers_changed_tot()	// Return number of layers with changes
-{
-	image_state *state;
+	image_info *image;
 	int j, k;
 
-	for (j = k =0; k <= layers_total; k++)	// Check each layer for mem_changed
+	for (j = k = 0; k <= layers_total; k++) // Check each layer for mem_changed
 	{
-		state = k == layer_selected ? &mem_state :
-			&layer_table[k].image->state_;
-		j += state->changed;
-		j += !state->filename[0];
+		image = k == layer_selected ? &mem_image :
+			&layer_table[k].image->image_;
+		j |= !!image->changed + !image->filename * 2;
 	}
 
-	return j;
+	return (j);
 }
 
-int check_layers_for_changes()			// 1=STOP, 2=IGNORE, 10=ESCAPE, -10=NOT CHECKED
+int check_layers_for_changes()		// 1=STOP, 2=IGNORE, -10=NOT CHANGED
 {
-	int i = -10, j = 0;
-	char *warning = _("One or more of the layers contains changes that have not been saved.  Do you really want to lose these changes?");
-
-
-	j = j + layers_changed_tot() + layers_changed;
-	if ( j>0 )
-		i = alert_box( _("Warning"), warning, _("Cancel Operation"), _("Lose Changes"), NULL );
-
-	return i;
+	if (!(layers_changed_tot() + layers_changed)) return (-10);
+	return (alert_box(_("Warning"),
+		_("One or more of the layers contains changes that have not been saved.  Do you really want to lose these changes?"),
+		_("Cancel Operation"), _("Lose Changes"), NULL));
 }
 
 static void layer_update_filename( char *name )
@@ -564,7 +552,7 @@ int load_layers( char *file_name )
 	layer_image *lim2;
 	char tin[300], load_name[PATHBUF], *c;
 	int i, j, k, kk, sens;
-	int layers_to_read = -1, layer_file_version = -1, lfail = 0, lplen = 0;
+	int layers_to_read = -1, /*layer_file_version = -1,*/ lfail = 0, lplen = 0;
 	FILE *fp;
 
 	c = strrchr(file_name, DIR_SEP);
@@ -580,7 +568,7 @@ int load_layers( char *file_name )
 
 	i = read_file_num(fp, tin);
 	if ( i==-987654321 ) goto fail2;
-	layer_file_version = i;
+//	layer_file_version = i;
 	if ( i>LAYERS_VERSION ) goto fail2;		// Version number must be compatible
 
 	i = read_file_num(fp, tin);
@@ -594,7 +582,7 @@ int load_layers( char *file_name )
 		// Read filename, strip end chars & try to load (if name length > 0)
 		fgets(tin, 256, fp);
 		string_chop(tin);
-		snprintf(load_name, PATHBUF, "%.*s%s", lplen, file_name, tin);
+		wjstrcat(load_name, PATHBUF, file_name, lplen, tin, NULL);
 		k = 1;
 		j = detect_image_format(load_name);
 		if ((j > 0) && (j != FT_NONE) && (j != FT_LAYERS1))
@@ -610,8 +598,8 @@ int load_layers( char *file_name )
 		/* Update image variables after load */
 		t = layer_table + layers_total;
 		lim2 = t->image;
-		strncpy(lim2->state_.filename, load_name, PATHBUF);
-		init_istate(&lim2->state_, &lim2->image_);
+		// !!! No old name so no fuss with saving it
+		lim2->image_.filename = strdup(load_name);
 
 		fgets(tin, 256, fp);
 		string_chop(tin);
@@ -625,20 +613,17 @@ int load_layers( char *file_name )
 
 		kk = read_file_num(fp, tin);
 		k = read_file_num(fp, tin);
-		lim2->state_.xpm_trans = kk <= 0 ? -1 : k < 0 ? 0 : k > 255 ? 255 : k;
+		lim2->image_.trans = kk <= 0 ? -1 : k < 0 ? 0 : k > 255 ? 255 : k;
 
 		k = read_file_num(fp, tin);
 		t->opacity = k < 1 ? 1 : k > 100 ? 100 : k;
 
-		layers_total++;
-	}
-	if (layers_total)
-	{
-		layer_copy_to_main(--layers_total);
-		layer_selected = layers_total;
+		init_istate(&lim2->state_, &lim2->image_);
+		if (!layers_total++) layer_copy_to_main(0); // Update mem_state
 	}
+	if (layers_total) layers_total--;
 	layer_refresh_list();
-	layer_select_slot(layer_selected);
+	layer_choose(layers_total);
 
 	/* Read in animation data - only if all layers loaded OK
 	 * (to do otherwise is likely to result in SIGSEGV) */
@@ -651,7 +636,7 @@ int load_layers( char *file_name )
 	if (lfail) /* There were failures */
 	{
 		snprintf(tin, 300, _("%d layers failed to load"), lfail);
-		alert_box( _("Error"), tin, _("OK"), NULL, NULL );
+		alert_box(_("Error"), tin, NULL);
 	}
 
 	return 1;		// Success
@@ -661,6 +646,131 @@ fail:
 	return -1;
 }
 
+int load_to_layers(char *file_name, int ftype, int ani_mode)
+{
+	char *buf, *tail;
+	image_frame *frm;
+	image_info *image;
+	image_state *state;
+	layer_node *t;
+	layer_image *lim;
+	frameset fset;
+	int anim = file_formats[ftype].flags & FF_ANIM;
+	int i, j, l, dx, dy, sens, res, res0, lname;
+
+
+	/* Create buffer for name mangling */
+	lname = strlen(file_name);
+	buf = malloc(lname + 64);
+	if (!buf) return (FILE_MEM_ERROR);
+	strcpy(buf, file_name);
+	tail = buf + lname;
+
+	/* Remove old layers, load new frames */
+	sens = layers_sensitive(FALSE);
+	if (layers_total) layers_free_all();	// Remove all current layers
+	res = res0 = load_frameset(&fset, ani_mode, file_name, FS_LAYER_LOAD, ftype);
+
+	if (!fset.cnt) /* Failure - we have no image */
+	{
+		if (res == FILE_LIB_ERROR) res = -1; // Failure is complete
+	}
+	else /* Got some frames - convert into layers */
+	{
+		l = 0; // Start from layer 0
+		if (anim) /* Animation */
+		{
+			/* Create an empty indexed background of 0th frame's size */
+			frm = fset.frames;
+			do_new_one(frm->width, frm->height, 256, mem_pal_def, 1, FALSE);
+			l = 1; // Frames start from layer 1
+		}
+
+		for (i = 0; i < fset.cnt; i++ , l++)
+		{
+			frm = fset.frames + i;
+
+			t = layer_table + l;
+			res = FILE_MEM_ERROR;
+			if (!l) // Layer 0 aka current image
+			{
+				if (mem_new(frm->width, frm->height, frm->bpp, 0))
+					break;
+				layer_copy_from_main(0);
+			}
+			else
+			{
+				if (!(t->image = alloc_layer(frm->width, frm->height,
+					frm->bpp, 0, NULL))) break;
+			}
+			res = res0;
+			lim = t->image;
+			t->visible = !anim;
+			t->opacity = 100;
+			image = &lim->image_; state = &lim->state_;
+
+			/* Move frame data to image */
+			memcpy(image->img, frm->img, sizeof(chanlist));
+			memset(frm->img, 0, sizeof(chanlist));
+			image->trans = frm->trans;
+			mem_pal_copy(image->pal, frm->pal ? frm->pal :
+				fset.pal ? fset.pal : mem_pal_def);
+			image->cols = frm->cols;
+			t->x = frm->x; t->y = frm->y;
+			update_undo(image);
+
+			/* Create a name for this frame */
+			sprintf(tail, ".%03d", i);
+			// !!! No old name so no fuss with saving it
+			image->filename = strdup(buf);
+
+			init_istate(state, image);
+			if (!l) layer_copy_to_main(0); // Update everything
+		}
+		layers_total = l ? l - 1 : 0;
+
+		if (anim)
+		{
+// !!! These legacy things need be replaced by a per-layer field
+			preserved_gif_delay = ani_gif_delay = fset.frames[0].delay;
+
+			/* Build animation cycle for these layers */
+			memset(ani_cycle_table, 0, sizeof(ani_cycle_table));
+			ani_frame1 = ani_cycle_table[0].frame0 = 1;
+			ani_frame2 = ani_cycle_table[0].frame1 = i - 1;
+			ani_cycle_table[0].len = i - 1;
+			for (j = 1; j < i; j++)
+				ani_cycle_table[0].layers[j - 1] = j;
+
+			/* Center the first frame over background */
+			dx = (mem_width - layer_table[1].image->image_.width) / 2;
+			dy = (mem_height - layer_table[1].image->image_.height) / 2;
+			for (j = 1; j < i; j++)
+			{
+				layer_table[j].x += dx;
+				layer_table[j].y += dy;
+			}
+
+			/* Display 1st layer in sequence */
+			layer_table[1].visible = TRUE;
+			layer_copy_from_main(0);
+			layer_copy_to_main(layer_selected = 1);
+		}
+		update_main_with_new_layer();
+	}
+	mem_free_frames(&fset);
+
+	layer_refresh_list();
+	layers_sensitive(sens); // This also selects current slot
+
+	/* Name change so that layers file would not overwrite the source */
+	strcpy(tail, ".txt");
+	layer_update_filename(buf);
+
+	free(buf);
+	return (res);
+}
+
 /* Convert absolute filename 'file' into one relative to prefix */
 static void parse_filename(char *dest, char *prefix, char *file, int len)
 {
@@ -677,7 +787,10 @@ static void parse_filename(char *dest, char *prefix, char *file, int len)
 		/* Count number of DIR_SEP encountered on and after point i in
 		 * 'prefix', add a '../' for each found */
 		for (k = i; k < len; k++)
-			if (prefix[k] == DIR_SEP) strnncat( dest, "../", PATHBUF);
+		{
+			if (prefix[k] == DIR_SEP)
+				strnncat(dest, ".." DIR_SEP_STR, PATHBUF);
+		}
 		/* nip backwards on 'file' from i to previous DIR_SEP or
 		 * beginning and ... */
 		for (k = i; (k >= 0) && (file[k] != DIR_SEP); k--);
@@ -705,10 +818,10 @@ int layer_save_composite(char *fname, ls_settings *settings)
 		settings->bpp = 3;
 		if (layer_selected) /* Set up background transparency */
 		{
-			res = layer_table[0].image->state_.xpm_trans;
+			res = image->trans;
 			settings->xpm_trans = res;
 			settings->rgb_trans = res < 0 ? -1 :
-				PNG_2_INT(layer_table[0].image->image_.pal[res]);
+				PNG_2_INT(image->pal[res]);
 
 		}
 		res = save_image(fname, settings);
@@ -735,8 +848,7 @@ void layer_add_composite()
 		view_render_rgb(lim->image_.img[CHN_IMAGE], 0, 0,
 			image->width, image->height, 1);
 		/* Copy background's transparency */
-		lim->state_.xpm_trans = layer_selected ?
-			layer_table[0].image->state_.xpm_trans : mem_xpm_trans;
+		lim->image_.trans = image->trans;
 		/* Activate the result */
 		layer_show_new();
 	}
@@ -763,10 +875,10 @@ int save_layers( char *file_name )
 	for ( i=0; i<=layers_total; i++ )
 	{
 		t = layer_table + i;
-		parse_filename(comp_name, file_name, t->image->state_.filename, l);
+		parse_filename(comp_name, file_name, t->image->image_.filename, l);
 		fprintf( fp, "%s\n", comp_name );
 
-		xpm = t->image->state_.xpm_trans;
+		xpm = t->image->image_.trans;
 		fprintf(fp, "%s\n%i\n%i\n%i\n%i\n%i\n%i\n", t->name,
 			t->visible, t->x, t->y, xpm >= 0, xpm, t->opacity);
 	}
@@ -781,7 +893,7 @@ int save_layers( char *file_name )
 fail:
 	c = gtkuncpy(NULL, layers_filename, 0);
 	msg = g_strdup_printf(_("Unable to save file: %s"), c);
-	alert_box(_("Error"), msg, _("OK"), NULL, NULL);
+	alert_box(_("Error"), msg, NULL);
 	g_free(msg);
 	g_free(c);
 
@@ -791,29 +903,24 @@ fail:
 
 int check_layers_all_saved()
 {
-	if ( layers_unsaved_tot() > 0 )
-	{
-		alert_box( _("Warning"), _("One or more of the image layers has not been saved.  You must save each image individually before saving the layers text file in order to load this composite image in the future."), _("OK"), NULL, NULL );
-		return 1;
-	}
-
-	return 0;
+	if (layers_changed_tot() < 2) return (0);
+	alert_box(_("Warning"), _("One or more of the image layers has not been saved.  You must save each image individually before saving the layers text file in order to load this composite image in the future."), NULL);
+	return (1);
 }
 
 void layer_press_save()
 {
-	if (!layers_filename[0]) file_selector( FS_LAYER_SAVE );
-	else
-	{
-		check_layers_all_saved();
-		save_layers( layers_filename );
-	}
+	if (!layers_filename[0]) file_selector(FS_LAYER_SAVE);
+	else if (!check_layers_all_saved()) save_layers(layers_filename);
 }
 
 void layer_press_remove_all()
 {
 	int i = check_layers_for_changes();
-	if (i < 0) i = alert_box( _("Warning"), _("Do you really want to delete all of the layers?"), _("No"), _("Yes"), NULL );
+
+	if (i < 0) i = alert_box(_("Warning"),
+		_("Do you really want to delete all of the layers?"),
+		_("No"), _("Yes"), NULL);
 	if (i != 2) return;
 
 	layers_free_all();
@@ -867,9 +974,8 @@ static void layer_inputs_changed(GtkObject *thing, gpointer user_data)
 		repaint_layer(layer_selected);
 		break;
 	case 3: // Transparency spin
-		mem_xpm_trans = gtk_spin_button_get_value_as_int(
-			GTK_SPIN_BUTTON(layer_spin));
-		update_stuff(UPD_TRANS);
+		mem_set_trans(gtk_spin_button_get_value_as_int(
+			GTK_SPIN_BUTTON(layer_spin)));
 		break;
 	}
 }
@@ -915,7 +1021,7 @@ static void layer_select(GtkList *list, GtkWidget *widget, gpointer user_data)
 
 	mt_spinslide_set_value(layer_slider, j ? t->opacity : 100);
 	layer_show_position();
-	gtk_spin_button_set_value(GTK_SPIN_BUTTON(layer_spin), j ? mem_xpm_trans : -1);
+	layer_show_trans();
 
 // !!! May cause list to be stuck in drag mode (release handled before press?)
 //	while (gtk_events_pending()) gtk_main_iteration();
@@ -927,15 +1033,18 @@ gboolean delete_layers_window()
 	// No deletion if no window, or inside a sensitive action
 	if (!layers_window || !layers_initialized) return (TRUE);
 	// Stop user prematurely exiting while drag 'n' drop loading
-	if (!GTK_WIDGET_SENSITIVE(layers_window)) return (TRUE);
+	if (!GTK_WIDGET_SENSITIVE(layers_box)) return (TRUE);
 
 	layers_initialized = FALSE;
 	win_store_pos(layers_window, "layers");
 
+	gtk_widget_hide(layers_window); // Butcher it when out of sight :-)
+	dock_undock(DOCK_LAYERS, TRUE);
 	gtk_widget_destroy(layers_window);
 	layers_window = NULL;
 	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(
 		menu_widgets[MENU_LAYER]), FALSE); // Ensure it's unchecked
+	layers_initialized = !!layers_box; // It may be docked now
 
 	return (FALSE);
 }
@@ -948,8 +1057,7 @@ void pressed_paste_layer()
 
 	if (layers_total >= MAX_LAYERS)
 	{
-		alert_box(_("Error"), _("You cannot add any more layers."),
-			_("OK"), NULL, NULL);
+		alert_box(_("Error"), _("You cannot add any more layers."), NULL);
 		return;
 	}
 
@@ -1011,7 +1119,7 @@ void pressed_paste_layer()
 		}
 	}
 
-	set_new_filename(layers_total, "");
+	set_new_filename(layers_total, NULL);
 
 	layer_show_new();
 //	pressed_layers();
@@ -1037,22 +1145,27 @@ void move_layer_relative(int l, int change_x, int change_y)	// Move a layer & up
 	}
 
 	layers_notify_changed();
-	if (l == layer_selected) layer_show_position();
+	if (l == layer_selected)
+	{
+		layer_show_position();
+		// All layers get moved while the current one stays still
+		if (show_layers_main) update_stuff(UPD_RENDER);
+	}
+	else if (show_layers_main) main_update_area(lx, ly, lw, lh);
 	vw_update_area(lx, ly, lw, lh);
-	if (show_layers_main) main_update_area(lx, ly, lw, lh);
 }
 
 #undef _
 #define _(X) X
 
 static toolbar_item layer_bar[] = {
-	{ _("New Layer"), -1, LTB_NEW, 0, xpm_new_xpm, ACT_LR_ADD, LR_NEW },
-	{ _("Raise"), -1, LTB_RAISE, 0, xpm_up_xpm, ACT_LR_SHIFT, 1 },
-	{ _("Lower"), -1, LTB_LOWER, 0, xpm_down_xpm, ACT_LR_SHIFT, -1 },
-	{ _("Duplicate Layer"), -1, LTB_DUP, 0, xpm_copy_xpm, ACT_LR_ADD, LR_DUP },
-	{ _("Centralise Layer"), -1, LTB_CENTER, 0, xpm_centre_xpm, ACT_LR_CENTER, 0 },
-	{ _("Delete Layer"), -1, LTB_DEL, 0, xpm_cut_xpm, ACT_LR_DEL, 0 },
-	{ _("Close Layers Window"), -1, LTB_CLOSE, 0, xpm_close_xpm, DLG_LAYERS, 1 },
+	{ _("New Layer"), -1, LTB_NEW, 0, XPM_ICON(new), ACT_LR_ADD, LR_NEW },
+	{ _("Raise"), -1, LTB_RAISE, 0, XPM_ICON(up), ACT_LR_SHIFT, 1 },
+	{ _("Lower"), -1, LTB_LOWER, 0, XPM_ICON(down), ACT_LR_SHIFT, -1 },
+	{ _("Duplicate Layer"), -1, LTB_DUP, 0, XPM_ICON(copy), ACT_LR_ADD, LR_DUP },
+	{ _("Centralise Layer"), -1, LTB_CENTER, 0, XPM_ICON(centre), ACT_LR_CENTER, 0 },
+	{ _("Delete Layer"), -1, LTB_DEL, 0, XPM_ICON(cut), ACT_LR_DEL, 0 },
+	{ _("Close Layers Window"), -1, LTB_CLOSE, 0, XPM_ICON(close), DLG_LAYERS, 1 },
 	{ NULL }};
 
 #undef _
@@ -1076,24 +1189,17 @@ static GtkWidget *layer_toolbar(GtkWidget **wlist)
 	return toolbar;
 }
 
-void pressed_layers()
+void create_layers_box()
 {
 	GtkWidget *vbox, *hbox, *table, *label, *tog, *scrolledwindow, *item;
-	GtkAccelGroup* ag;
 	char txt[32];
 	int i;
 
 
-	if (layers_window) return; // Already have one
 	layers_initialized = FALSE;
-
-	layers_window = add_a_window( GTK_WINDOW_TOPLEVEL, "", GTK_WIN_POS_NONE, FALSE );
-	win_restore_pos(layers_window, "layers", 0, 0, 400, 400);
-
-	vbox = gtk_vbox_new (FALSE, 0);
-	gtk_widget_show (vbox);
-	gtk_container_add (GTK_CONTAINER (layers_window), vbox);
-	gtk_container_set_border_width (GTK_CONTAINER (vbox), 5);
+	layers_box = vbox = gtk_vbox_new(FALSE, 0);
+	gtk_widget_show(vbox);
+	gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
 
 	scrolledwindow = xpack(vbox, gtk_scrolled_window_new(NULL, NULL));
 	gtk_widget_show (scrolledwindow);
@@ -1110,7 +1216,7 @@ void pressed_layers()
 	gtk_container_set_focus_vadjustment(GTK_CONTAINER(layer_list),
 		gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(scrolledwindow)));
 
-	for ( i=MAX_LAYERS; i>=0; i-- )
+	for (i = MAX_LAYERS; i >= 0; i--)
 	{
 		hbox = gtk_hbox_new(FALSE, 3);
 
@@ -1165,13 +1271,16 @@ void pressed_layers()
 	add_to_table( _("Layer Name"), table, 0, 0, 0 );
 	add_to_table( _("Position"), table, 1, 0, 0 );
 	add_to_table( _("Opacity"), table, 2, 0, 0 );
-	add_to_table( _("Transparent Colour"), table, 3, 0, 0 );
+
+	label = gtk_label_new(_("Transparent Colour"));
+	gtk_widget_show(label);
+	to_table_l(label, table, 3, 0, 2, 0);
+	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
 
 	entry_layer_name = gtk_entry_new_with_max_length(LAYER_NAMELEN - 1);
 	gtk_widget_set_usize(entry_layer_name, 100, -2);
-	gtk_widget_show (entry_layer_name);
-	gtk_table_attach (GTK_TABLE (table), entry_layer_name, 1, 3, 0, 1,
-		(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
+	gtk_widget_show(entry_layer_name);
+	to_table_l(entry_layer_name, table, 0, 1, 2, 0);
 	gtk_signal_connect(GTK_OBJECT(entry_layer_name), "changed",
 		GTK_SIGNAL_FUNC(layer_inputs_changed), (gpointer)0);
 
@@ -1188,25 +1297,48 @@ void pressed_layers()
 		(gpointer)2);
 	mt_spinslide_set_value(layer_slider, layer_table[layer_selected].opacity);
 
-	layer_spin = spin_to_table(table, 3, 1, 0, 0, -1, 255);
+	layer_spin = spin_to_table(table, 3, 2, 0, 0, -1, 255);
 	spin_connect(layer_spin, GTK_SIGNAL_FUNC(layer_inputs_changed), (gpointer)3);
 
 	pack(vbox, sig_toggle(_("Show all layers in main window"),
 		show_layers_main, NULL, GTK_SIGNAL_FUNC(layer_main_toggled)));
 
-	ag = gtk_accel_group_new();
-	gtk_widget_add_accelerator(layer_tools[LTB_CLOSE], "clicked", ag,
-		GDK_Escape, 0, (GtkAccelFlags) 0);
-
-	gtk_signal_connect_object (GTK_OBJECT (layers_window), "delete_event",
-		GTK_SIGNAL_FUNC (delete_layers_window), NULL);
-
-	layers_update_titlebar();
-
 	/* !!! Select *before* show - otherwise it's nontrivial (try & see) */
 	layers_initialized = TRUE;
 	layer_select_slot(layer_selected);
-	gtk_window_set_transient_for( GTK_WINDOW(layers_window), GTK_WINDOW(main_window) );
+
+	/* Make widget to report its demise */
+	gtk_signal_connect(GTK_OBJECT(layers_box), "destroy",
+		GTK_SIGNAL_FUNC(gtk_widget_destroyed), &layers_box);
+
+	/* Keep the invariant */
+	gtk_widget_ref(layers_box);
+	gtk_object_sink(GTK_OBJECT(layers_box));
+}
+
+void pressed_layers()
+{
+	GtkAccelGroup *ag;
+
+
+	if (layers_window) return; // Already have it open
+
+	layers_window = add_a_window(GTK_WINDOW_TOPLEVEL, "", GTK_WIN_POS_NONE, FALSE);
+	win_restore_pos(layers_window, "layers", 0, 0, 400, 400);
+	layers_update_titlebar();
+
+	dock_undock(DOCK_LAYERS, FALSE);
+	gtk_container_add(GTK_CONTAINER(layers_window), layers_box);
+	gtk_widget_unref(layers_box);
+
+	ag = gtk_accel_group_new();
+	gtk_widget_add_accelerator(layer_tools[LTB_CLOSE], "clicked", ag,
+		GDK_Escape, 0, (GtkAccelFlags)0);
+
+	gtk_signal_connect_object(GTK_OBJECT(layers_window), "delete_event",
+		GTK_SIGNAL_FUNC(delete_layers_window), NULL);
+
+	gtk_window_set_transient_for(GTK_WINDOW(layers_window), GTK_WINDOW(main_window));
 	gtk_widget_show(layers_window);
-	gtk_window_add_accel_group(GTK_WINDOW (layers_window), ag);
+	gtk_window_add_accel_group(GTK_WINDOW(layers_window), ag);
 }
