@@ -1,5 +1,5 @@
 /*	layer.h
-	Copyright (C) 2005-2008 Mark Tyler and Dmitry Groshev
+	Copyright (C) 2005-2011 Mark Tyler and Dmitry Groshev
 
 	This file is part of mtPaint.
 
@@ -48,7 +48,7 @@ typedef struct
 } layer_node;
 
 
-GtkWidget *layers_window;
+GtkWidget *layers_box, *layers_window;
 
 layer_node layer_table[MAX_LAYERS+1];	// Table of layer info
 
@@ -65,16 +65,19 @@ int	show_layers_main,		// Show all layers in main window
 ///	PROCEDURES
 
 void layers_init();
-layer_image *alloc_layer(int w, int h, int bpp, int cmask, chanlist src);
+layer_image *alloc_layer(int w, int h, int bpp, int cmask, image_info *src);
 void pressed_layers();
 void pressed_paste_layer();
 gboolean delete_layers_window();
+void create_layers_box();
 
 int load_layers( char *file_name );
 int save_layers( char *file_name );
 void layer_press_save();
 int layer_save_composite(char *fname, ls_settings *settings);
 
+int load_to_layers(char *file_name, int ftype, int ani_mode);
+
 #define update_main_with_new_layer() update_stuff(UPD_LAYER)
 
 void layers_notify_changed();
@@ -99,3 +102,5 @@ void layer_press_delete();		// Delete current layer
 
 void string_chop( char *txt );
 int read_file_num(FILE *fp, char *txt);
+
+void layer_show_trans();		// Update transparency in layers window
