@@ -1,5 +1,5 @@
 /*	memory.h
-	Copyright (C) 2004-2009 Mark Tyler and Dmitry Groshev
+	Copyright (C) 2004-2011 Mark Tyler and Dmitry Groshev
 
 	This file is part of mtPaint.
 
@@ -17,6 +17,7 @@
 	along with mtPaint in the file COPYING.
 */
 
+#include <limits.h>
 #include <png.h>
 
 /// Definitions, structures & variables
@@ -27,6 +28,7 @@
 #define MIN_HEIGHT 1
 /* !!! If MAX_WIDTH * MAX_HEIGHT * max bpp won't fit into int, lots of code
  * !!! will have to be modified to use size_t instead */
+#define MAX_DIM (MAX_WIDTH > MAX_HEIGHT ? MAX_WIDTH : MAX_HEIGHT)
 
 #define DEFAULT_WIDTH 640
 #define DEFAULT_HEIGHT 480
@@ -93,14 +95,57 @@
 
 #define SIZEOF_PALETTE (256 * sizeof(png_color))
 
+// Both limits should be powers of two
+#define FRAMES_MIN 16
+#define FRAMES_MAX (1024 * 1024) /* A million frames should be QUITE enough */
+
+/* Frame flags */
+#define FM_DISPOSAL     3 /* Disposal mask */
+#define FM_DISP_REMOVE  0 /* Remove (make transparent) (default) */
+#define FM_DISP_LEAVE   1 /* Leave in place */
+#define FM_DISP_RESTORE 2 /* Restore to previous state */
+#define FM_NUKE         4 /* Delete this frame at earliest opportunity */
+
+/* Undo data types */
+#define UD_FILENAME  0 /* Filename */
+#define UD_TEMPFILES 1 /* Temp files list */
+#define NUM_UTYPES   2 /* Should be no larger than 32 */
+//	List in here all types which need freeing
+#define UD_FREE_MASK (1 << UD_FILENAME)
+
 typedef unsigned char *chanlist[NUM_CHANNELS];
 
+typedef struct {
+	chanlist img;
+	png_color *pal;
+	int width, height;
+	int x, y, delay;
+	short cols, bpp, trans;
+	unsigned short flags;
+} image_frame;
+
+typedef struct {
+	image_frame *frames;	// Pointer to frames array
+	png_color *pal;		// Default palette
+	int cur;		// Index of current frame
+	int cnt;		// Number of frames in use
+	int max;		// Total number of frame slots
+	size_t size;		// Total used memory (0 means count it anew)
+} frameset;
+
+typedef struct {
+	unsigned int map;
+	void *store[NUM_UTYPES];
+} undo_data;
+
 typedef struct {
 	chanlist img;
 	png_color *pal_;
 	unsigned char *tileptr;
-	int cols, width, height, bpp, flags;
+	undo_data *dataptr;
 	size_t size;
+	int width, height, flags;
+	short cols, bpp, trans;
 } undo_item;
 
 typedef struct {
@@ -117,19 +162,19 @@ typedef struct {
 	png_color pal[256];	// RGB entries for all 256 palette colours
 	int cols;	// Number of colours in the palette: 1..256 or 0 for no image
 	int bpp;		// Bytes per pixel = 1 or 3
+	int trans;		// Transparent colour index (-1 if none)
 	int width, height;	// Image geometry
 	undo_stack undo_;	// Image's undo stack
+	char *filename;		// File name of file loaded/saved
+	void *tempfiles;	// List of up-to-date temp files
+	int changed;		// Changed since last load/save flag
 } image_info;
 
 typedef struct {
-	char filename[PATHBUF];		// File name of file loaded/saved
-	char *tempname;			// File name of up-to-date temp file
 	int channel;			// Current active channel
-	int changed;			// Changed since last load/save flag
 	int ics;			// Has the centre been set by the user?
-	float icx, icy;			// Current centre x,y
+	float ic[2];			// Current centre x,y
 	int tool_pat;			// Tool pattern number
-	int xpm_trans;			// Transparent colour index (-1 if none)
 	int xbm_hot_x, xbm_hot_y;	// Current XBM hot spot
 	char prot_mask[256];		// 256 bytes used for indexed images
 	int prot;			// Number of protected colours in prot_RGB
@@ -145,7 +190,7 @@ typedef struct {
 
 typedef struct {
 	/* Base values */
-	int status, x1, y1, x2, y2;	// Gradient placement tool
+	int status, xy[4];	// Gradient placement tool
 	int len, rep, ofs;	// Gradient length, repeat, and offset
 	int gmode, rmode;	// Gradient mode and repeat mode
 	/* Derived values */
@@ -190,10 +235,13 @@ int grad_opacity;			// Preview opacity
 
 /* Gradient types */
 #define GRAD_TYPE_RGB      0
-#define GRAD_TYPE_HSV      1
-#define GRAD_TYPE_BK_HSV   2
-#define GRAD_TYPE_CONST    3
-#define GRAD_TYPE_CUSTOM   4
+#define GRAD_TYPE_SRGB     1
+#define GRAD_TYPE_HSV      2
+#define GRAD_TYPE_BK_HSV   3
+#define GRAD_TYPE_CONST    4
+#define GRAD_TYPE_CUSTOM   5
+
+#define STROKE_GRADIENT (mem_gradient && (gradient[mem_channel].status == GRAD_NONE))
 
 /// Bayer ordered dithering
 
@@ -255,6 +303,7 @@ double flood_step;
 int flood_cube, flood_img, flood_slide;
 
 int smudge_mode;
+int posterize_mode;	// bitwise/truncated/rounded
 
 /// QUANTIZATION SETTINGS
 
@@ -274,6 +323,7 @@ image_info mem_image;			// Current image
 #define mem_pal		mem_image.pal
 #define mem_cols	mem_image.cols
 #define mem_img_bpp	mem_image.bpp
+#define mem_xpm_trans	mem_image.trans
 #define mem_width	mem_image.width
 #define mem_height	mem_image.height
 
@@ -283,6 +333,10 @@ image_info mem_image;			// Current image
 #define mem_undo_redo		mem_image.undo_.redo
 #define mem_undo_max		mem_image.undo_.max
 
+#define mem_filename		mem_image.filename
+#define mem_tempfiles		mem_image.tempfiles
+#define mem_changed		mem_image.changed
+
 image_info mem_clip;			// Current clipboard
 
 #define mem_clipboard		mem_clip.img[CHN_IMAGE]
@@ -300,18 +354,18 @@ image_info mem_clip;			// Current clipboard
 #define mem_clip_real_w		mem_clip.undo_.items[OLD_CLIP].width
 #define mem_clip_real_h		mem_clip.undo_.items[OLD_CLIP].height
 #define mem_clip_real_clear()	mem_free_image(&mem_clip, FREE_UNDO)
+// Repurpose the field
+#define mem_clip_paletted	mem_clip.changed
+#define mem_clip_pal		mem_clip.pal
 
 image_state mem_state;			// Current edit settings
 
-#define mem_filename		mem_state.filename
-#define mem_tempname		mem_state.tempname
 #define mem_channel		mem_state.channel
-#define mem_changed		mem_state.changed
-#define mem_icx			mem_state.icx
-#define mem_icy			mem_state.icy
+#define mem_ic			mem_state.ic
+#define mem_icx			mem_state.ic[0]
+#define mem_icy			mem_state.ic[1]
 #define mem_ics			mem_state.ics
 #define mem_tool_pat		mem_state.tool_pat
-#define mem_xpm_trans		mem_state.xpm_trans
 #define mem_xbm_hot_x		mem_state.xbm_hot_x
 #define mem_xbm_hot_y		mem_state.xbm_hot_y
 #define mem_prot_mask		mem_state.prot_mask
@@ -327,7 +381,6 @@ image_state mem_state;			// Current edit settings
 int mem_clip_x, mem_clip_y;		// Clipboard location on canvas
 
 extern unsigned char mem_brushes[];	// Preset brushes image
-int brush_tool_type;			// Last brush tool type
 int mem_brush_list[81][3];		// Preset brushes parameters
 int mem_nudge;				// Nudge pixels per SHIFT+Arrow key during selection/paste
 
@@ -343,14 +396,27 @@ unsigned char mem_pattern[8 * 8];	// Current pattern
 unsigned char mem_col_pat[8 * 8];	// Indexed 8x8 colourised pattern using colours A & B
 unsigned char mem_col_pat24[8 * 8 * 3];	// RGB 8x8 colourised pattern using colours A & B
 
-/// PREVIEW/TOOLS
+/// TOOLS
+
+typedef struct {
+	int type, brush;
+	int var[3];		// Size, flow, opacity
+} tool_info;
+
+tool_info tool_state;
+
+#define tool_type	tool_state.type		/* Currently selected tool */
+#define brush_type	tool_state.brush	/* Last brush tool type */
+#define tool_size	tool_state.var[0]
+#define tool_flow	tool_state.var[1]
+#define tool_opacity	tool_state.var[2]	/* Opacity - 255 = solid */
 
-int tool_type, tool_size, tool_flow;	// Currently selected tool
-int tool_opacity;			// Opacity - 255 = solid
 int pen_down;				// Are we drawing? - Used to see if we need to do an UNDO
 int tool_ox, tool_oy;			// Previous tool coords - used by continuous mode
 int mem_continuous;			// Area we painting the static shapes continuously?
 
+/// PREVIEW
+
 int mem_brcosa_allow[3];		// BRCOSA RGB
 
 /// PALETTE
@@ -362,6 +428,39 @@ extern unsigned char mem_pals[];	// RGB screen memory holding current palette
 int mem_background;			// Non paintable area
 int mem_histogram[256];
 
+/// Next power of two
+
+static inline unsigned int nextpow2(unsigned int n)
+{
+	n |= n >> 1; n |= n >> 2; n |= n >> 4; n |= n >> 8; n |= n >> 16;
+#if UINT_MAX > 0xFFFFFFFFUL
+	n |= n >> 32;
+#endif
+	return (n + 1);
+}
+
+// Number of set bits
+
+static inline unsigned int bitcount(unsigned int n)
+{
+	unsigned int m;
+#if UINT_MAX > 0xFFFFFFFFUL
+	n -= (n >> 1) & 0x5555555555555555ULL;
+	m = n & 0xCCCCCCCCCCCCCCCCULL; n = (n ^ m) + (m >> 2);
+	n = (n + (n >> 4)) & 0x0F0F0F0F0F0F0F0FULL;
+	n = (n + (n >> 8)) & 0x00FF00FF00FF00FFULL;
+	n = (n + (n >> 16)) & 0x0000FFFF0000FFFFULL;
+	n = (n + (n >> 32)) & 0x00000000FFFFFFFFULL;
+#else
+	n -= (n >> 1) & 0x55555555;
+	m = n & 0x33333333; n = m + ((n ^ m) >> 2);
+	n = (n + (n >> 4)) & 0x0F0F0F0F;
+	n = (n + (n >> 8)) & 0x00FF00FF;
+	n = (n + (n >> 16)) & 0x0000FFFF;
+#endif
+	return (n);
+}
+
 /// Floored integer division
 
 static inline int floor_div(int dd, int dr)
@@ -369,16 +468,50 @@ static inline int floor_div(int dd, int dr)
 	return (dd / dr - (dd % dr < 0)); // optimizes to perfection on x86
 }
 
+/// Table-based translation
+
+static inline void do_xlate(unsigned char *xlat, unsigned char *data, int len)
+{
+	int i;
+
+	for (i = 0; i < len; i++) data[i] = xlat[data[i]];
+}
+
+/// Copying "x0y0x1y1" quad
+
+#define copy4(D,S) memcpy(D, S, 4 * sizeof(int))
+
+/// Block allocator
+
+typedef struct {
+	char *block;
+	unsigned int here, size;
+	unsigned int minsize, incr;
+} wjmem;
+
+wjmem *wjmemnew(int size, int incr);
+void wjmemfree(wjmem *mem);
+void *wjmalloc(wjmem *mem, int size, int align);
+
 /// Multiblock allocator
 
-void *multialloc(int align, void *ptr, int size, ...);
+#define MA_ALIGN_MASK    0x03 /* For alignment constraints */
+#define MA_ALIGN_DEFAULT 0x00
+#define MA_ALIGN_DOUBLE  0x01
+#define MA_SKIP_ZEROSIZE 0x04 /* Leave pointers to zero-size areas unchanged */
 
-/// Vectorized low-level drawing function
+void *multialloc(int flags, void *ptr, int size, ...);
+
+/// Vectorized low-level drawing functions
 
 void (*put_pixel)(int x, int y);
+void (*put_pixel_row)(int x, int y, int len, unsigned char *xsel);
+
+	// Intersect two rectangles
+int clip(int *rxy, int x0, int y0, int x1, int y1, const int *vxy);
 
 	// Intersect outer & inner rectangle, write out what it separates into
-int clip4(int *xywh04, int xo, int yo, int wo, int ho, int xi, int yi, int wi, int hi);
+int clip4(int *rect04, int xo, int yo, int wo, int ho, int xi, int yi, int wi, int hi);
 
 /// Line iterator
 
@@ -388,11 +521,21 @@ typedef int linedata[10];
 void line_init(linedata line, int x0, int y0, int x1, int y1);
 int line_step(linedata line);
 void line_nudge(linedata line, int x, int y);
-int line_clip(linedata line, int *vxy, int *step);
+int line_clip(linedata line, const int *vxy, int *step);
+void line_flip(linedata line);
 
 /// Procedures
 
+//	Add one more frame to a frameset
+int mem_add_frame(frameset *fset, int w, int h, int bpp, int cmask, png_color *pal);
+//	Remove specified frame from a frameset
+void mem_remove_frame(frameset *fset, int frame);
+//	Empty a frameset
+void mem_free_frames(frameset *fset);
+
 void init_istate(image_state *state, image_info *image);	// Set initial state of image variables
+void mem_replace_filename(int layer, char *fname);	// Change layer's filename
+void mem_file_modified(char *fname);	// Label file's frames in current layer as changed
 int init_undo(undo_stack *ustack, int depth);	// Create new undo stack of a given depth
 void update_undo_depth();	// Resize all undo stacks
 
@@ -407,10 +550,6 @@ int mem_cols_used_real(unsigned char *im, int w, int h, int max_count, int prog)
 			// Count colours used in RGB chunk and dump to found table
 void mem_cols_found(png_color *userpal);	// Convert colours list into palette
 
-
-int read_hex( char in );			// Convert character to hex value 0..15.  -1=error
-int read_hex_dub( char *in );			// Read hex double
-
 #define FREE_IMAGE 1
 #define FREE_UNDO  2
 #define FREE_ALL   3
@@ -424,7 +563,7 @@ void mem_free_image(image_info *image, int mode);
 
 //	Allocate new image data
 int mem_alloc_image(int mode, image_info *image, int w, int h, int bpp,
-	int cmask, chanlist src);
+	int cmask, image_info *src);
 //	Allocate space for new image, removing old if needed
 int mem_new( int width, int height, int bpp, int cmask );
 //	Allocate new clipboard, removing or preserving old as needed
@@ -458,17 +597,29 @@ void mem_bacteria( int val );			// Apply bacteria effect val times the canvas ar
 void mem_gauss(double radiusX, double radiusY, int gcor);
 void mem_unsharp(double radius, double amount, int threshold, int gcor);
 void mem_dog(double radiusW, double radiusN, int norm, int gcor);
-void mem_kuwahara(int r, int gcor);
+void mem_kuwahara(int r, int gcor, int detail);
 
+/* Colorspaces */
+#define CSPACE_RGB  0
+#define CSPACE_SRGB 1
+#define CSPACE_LXN  2
+#define NUM_CSPACES 3
+
+/* Distance measures */
+#define DIST_LINF     0 /* Largest absolute difference (Linf measure) */
+#define DIST_L1       1 /* Sum of absolute differences (L1 measure) */
+#define DIST_L2       2 /* Euclidean distance (L2 measure) */
+#define NUM_DISTANCES 3
 
 /// PALETTE PROCS
 
-int mem_load_pal( char *file_name, png_color *pal );	// Load file into palette array >1 => cols read
 void mem_pal_load_def();		// Load default palette
 
 #define mem_pal_copy(A, B) memcpy((A), (B), SIZEOF_PALETTE)
 void mem_pal_init();			// Initialise whole of palette RGB
 void mem_greyscale(int gcor);		// Convert image to greyscale
+void do_convert_rgb(int start, int step, int cnt, unsigned char *dest,
+	unsigned char *src, png_color *pal);
 int mem_convert_rgb();			// Convert image to RGB
 int mem_convert_indexed();		// Convert image to Indexed Palette
 //	Quantize image using Max-Min algorithm
@@ -494,13 +645,13 @@ void mem_mask_init();			// Initialise RGB protection mask
 int mem_protected_RGB(int intcol);	// Is this intcol in list?
 
 void mem_swap_cols(int redraw);		// Swaps colours and update memory
+void mem_set_trans(int trans);		// Set transparent colour and update
 void mem_get_histogram(int channel);	// Calculate how many of each colour index is on the canvas
 int scan_duplicates();			// Find duplicate palette colours
 void remove_duplicates();		// Remove duplicate palette colours - call AFTER scan_duplicates
 int mem_remove_unused_check();		// Check to see if we can remove unused palette colours
 int mem_remove_unused();		// Remove unused palette colours
-void mem_scale_pal(png_color *pal, int i1, int r1, int g1, int b1,
-	int i2, int r2, int g2, int b2); // Generate a scaled palette
+void mem_bw_pal(png_color *pal, int i1, int i2); // Generate black-to-white palette
 //	Create colour-transformed palette
 void transform_pal(png_color *pal1, png_color *pal2, int p1, int p2);
 void mem_pal_sort( int a, int i1, int i2, int rev );
@@ -511,23 +662,26 @@ void mem_canvas_index_move( int c1, int c2 );	// Similar to palette item move bu
 
 void set_zoom_centre( int x, int y );
 
-// Nonclassical HSV: H is 0..6, S is 0.. 1, V is 0..255
+// Nonclassical HSV: H is 0..6, S is 0..1, V is 0..255
 void rgb2hsv(unsigned char *rgb, double *hsv);
 void hsv2rgb(unsigned char *rgb, double *hsv);
 
 //// UNDO
 
-#define UNDO_PAL   0	/* Palette changes */
-#define UNDO_XPAL  1	/* Palette and indexed image changes */
-#define UNDO_COL   2	/* Palette and/or RGB image changes */
-#define UNDO_DRAW  3	/* Changes to current channel / RGBA */
-#define UNDO_INV   4	/* "Invert" operation */
-#define UNDO_XFORM 5	/* Changes to all channels */
-#define UNDO_FILT  6	/* Changes to current channel */
-#define UNDO_PASTE 7	/* Paste operation (current / RGBA) */
-#define UNDO_TOOL  8	/* Same as UNDO_DRAW but respects pen_down */
+enum {
+	UNDO_PAL = 0,	/* Palette changes */
+	UNDO_XPAL,	/* Palette and indexed image changes */
+	UNDO_COL,	/* Palette and/or RGB image changes */
+	UNDO_DRAW,	/* Changes to current channel / RGBA */
+	UNDO_INV,	/* "Invert" operation */
+	UNDO_XFORM,	/* Changes to all channels */
+	UNDO_FILT,	/* Changes to current channel */
+	UNDO_PASTE,	/* Paste operation (current / RGBA) */
+	UNDO_TOOL,	/* Same as UNDO_DRAW but respects pen_down */
+	UNDO_TRANS	/* Transparent colour change (cumulative) */
+};
 
-int mem_undo_next(int mode);		// Call this after a draw event but before any changes to image
+void mem_undo_next(int mode);	// Call this after a draw event but before any changes to image
 //	 Get address of previous channel data (or current if none)
 unsigned char *mem_undo_previous(int channel);
 void mem_undo_prepare();	// Call this after changes to image, to compress last frame
@@ -540,6 +694,7 @@ void mem_undo_forward();		// REDO requested by user
 #define UC_DELETE  0x04	/* Force delete */
 #define UC_PENDOWN 0x08	/* Respect pen_down */
 #define UC_GETMEM  0x10 /* Get memory and do nothing */
+#define UC_ACCUM   0x20 /* Cumulative change */
 
 int undo_next_core(int mode, int new_width, int new_height, int new_bpp, int cmask);
 void update_undo(image_info *image);	// Copy image state into current undo frame
@@ -548,9 +703,9 @@ void *mem_try_malloc(size_t size);
 
 //// Drawing Primitives
 
-int sb_xywh[4];				// Backbuffer placement
+int sb_rect[4];				// Backbuffer placement
 int init_sb();				// Create shapeburst backbuffer
-void render_sb();			// Render from shapeburst backbuffer
+void render_sb(unsigned char *mask);	// Render from shapeburst backbuffer
 
 int mem_clip_mask_init(unsigned char val);		// Initialise the clipboard mask
 //	Extract alpha info from RGB clipboard
@@ -576,16 +731,22 @@ void mem_flip_h( char *mem, int w, int h, int bpp );		// Flip image horizontally
 int mem_sel_rot( int dir );					// Rotate clipboard 90 degrees
 int mem_image_rot( int dir );					// Rotate canvas 90 degrees
 
+//	Get new image geometry of rotation. angle = degrees
 void mem_rotate_geometry(int ow, int oh, double angle, int *nw, int *nh);
-				// Get new image geometry of rotation. angle = degrees
-// Rotate canvas or clipboard by any angle (degrees)
+//	Rotate canvas or clipboard by any angle (degrees)
 int mem_rotate_free(double angle, int type, int gcor, int clipboard);
 void mem_rotate_free_real(chanlist old_img, chanlist new_img, int ow, int oh,
 	int nw, int nh, int bpp, double angle, int mode, int gcor, int dis_a,
 	int silent);
 
-int mem_image_scale(int nw, int nh, int type, int gcor, int sharp);	// Scale image
-int mem_image_scale_real(chanlist old_img, int ow, int oh, int bpp, chanlist new_img, int nw, int nh, int type, int gcor, int sharp);
+#define BOUND_MIRROR 0 /* Mirror image beyond edges */
+#define BOUND_TILE   1 /* Tiled image beyond edges */
+#define BOUND_VOID   2 /* Transparency beyond edges */
+
+//	Scale image
+int mem_image_scale(int nw, int nh, int type, int gcor, int sharp, int bound);
+int mem_image_scale_real(chanlist old_img, int ow, int oh, int bpp,
+	chanlist new_img, int nw, int nh, int type, int gcor, int sharp);
 int mem_image_resize(int nw, int nh, int ox, int oy, int mode);	// Resize image
 
 int mem_isometrics(int type);
@@ -620,8 +781,10 @@ void draw_quad(linedata line1, linedata line2, linedata line3, linedata line4);
 #define BPP(x) ((x) == CHN_IMAGE ? mem_img_bpp : 1)
 #define IS_INDEXED ((mem_channel == CHN_IMAGE) && (mem_img_bpp == 1))
 
-#define POSTERIZE_MACRO res = 0.49 + ( ((1 << posty) - 1) * ((float) res)/255);\
-			res = 0.49 + 255 * ( ((float) res) / ((1 << posty) - 1) );
+/* Whether process_img() needs extra buffer passed in */
+#define NEED_XBUF_DRAW (mem_blend && (blend_mode & BLEND_MMASK))
+#define NEED_XBUF_PASTE (NEED_XBUF_DRAW || \
+	((mem_channel == CHN_IMAGE) && (mem_clip_bpp < mem_img_bpp)))
 
 void prep_mask(int start, int step, int cnt, unsigned char *mask,
 	unsigned char *mask0, unsigned char *img0);
@@ -630,9 +793,7 @@ void process_mask(int start, int step, int cnt, unsigned char *mask,
 	unsigned char *trans, int opacity, int noalpha);
 void process_img(int start, int step, int cnt, unsigned char *mask,
 	unsigned char *imgr, unsigned char *img0, unsigned char *img,
-	int sourcebpp, int destbpp);
-void paste_pixels(int x, int y, int len, unsigned char *mask, unsigned char *img,
-	unsigned char *alpha, unsigned char *trans, int opacity);
+	unsigned char *xbuf, int bpp, int opacity); // opacity value is ignored
 void copy_area(image_info *dest, image_info *src, int x, int y);
 
 // Retroactive masking - by blending with undo frame
@@ -641,17 +802,17 @@ void mask_merge(unsigned char *old, int channel, unsigned char *mask);
 int pixel_protected(int x, int y);				// generic
 void row_protected(int x, int y, int len, unsigned char *mask);
 void put_pixel_def( int x, int y );				// generic
+void put_pixel_row_def(int x, int y, int len, unsigned char *xsel); // generic
 int get_pixel( int x, int y );					// generic
 int get_pixel_RGB( int x, int y );				// converter
 int get_pixel_img( int x, int y );				// from image
 
 int grad_value(int *dest, int slot, double x);
-int grad_pixel(unsigned char *dest, int x, int y);
+void grad_pixels(int start, int step, int cnt, int x, int y, unsigned char *mask,
+	unsigned char *op0, unsigned char *img0, unsigned char *alpha0);
 void grad_update(grad_info *grad);
 void gmap_setup(grad_map *gmap, grad_store gstore, int slot);
-void grad_def_update();
-void prep_grad(int start, int step, int cnt, int x, int y, unsigned char *mask,
-	unsigned char *op0, unsigned char *img0, unsigned char *alpha0);
+void grad_def_update(int slot);
 
 #define GRAD_CUSTOM_DATA(X) ((X) ? GRAD_POINTS * ((X) * 4 + 2) : 0)
 #define GRAD_CUSTOM_DMAP(X) (GRAD_POINTS * ((X) * 4 + 3))
@@ -662,10 +823,50 @@ void blend_indexed(int start, int step, int cnt, unsigned char *rgb,
 	unsigned char *img0, unsigned char *img,
 	unsigned char *alpha0, unsigned char *alpha, int opacity);
 
+//	Select colors nearest to A->B gradient
+int mem_pick_gradient(unsigned char *buf, int cspace, int mode);
+
 int mem_skew(double xskew, double yskew, int type, int gcor);
 
 int average_pixels(unsigned char *rgb, int iw, int ih, int x, int y, int w, int h);
 
+//// SEGMENTATION
+
+#define SEG_PROGRESS 1
+
+typedef struct {
+	int which; // Pixel index * 2 + (0 if right / 1 if down)
+	float diff;
+} seg_edge;
+
+typedef struct {
+	unsigned int group, cnt;
+	unsigned char rank; // Value is logarithmic, so this is more than enough
+//	unsigned char reserved[3];
+	float threshold;
+} seg_pixel;
+
+typedef struct {
+	/* Working set */
+	seg_edge *edges;
+	seg_pixel *pix;
+	int w, h, cnt;
+	int phase; // Which phase is currently valid
+	/* Parameters */
+	int minrank;
+	int minsize;
+	double threshold;
+} seg_state;
+
+seg_state *mem_seg_prepare(seg_state *s, unsigned char *img, int w, int h,
+	int flags, int cspace, int dist);
+int mem_seg_process_chunk(int start, int cnt, seg_state *s);
+int mem_seg_process(seg_state *s);
+double mem_seg_threshold(seg_state *s);
+void mem_seg_scan(unsigned char *dest, int y, int x, int w, int zoom,
+	const seg_state *s);
+void mem_seg_render(unsigned char *img, const seg_state *s);
+
 #define IF_IN_RANGE( x, y ) if ( x>=0 && y>=0 && x<mem_width && y<mem_height )
 
 #define mtMIN(a,b,c) if ( b<c ) a=b; else a=c;
@@ -676,9 +877,23 @@ int average_pixels(unsigned char *rgb, int iw, int ih, int x, int y, int w, int
  * alignment; this macro serves as part of a workaround for that problem
  */
 #ifdef WIN32
-#define ALIGNTO(p,s) ((void *)(((int)(p) + sizeof(s) - 1) & (-sizeof(s))))
+/* In Win32, pointers fit into ints */
+#define ALIGN(p) ((void *)(((int)(p) + sizeof(double) - 1) & (-sizeof(double))))
 #else
-#define ALIGNTO(p,s) ((void *)(p))
+#define ALIGN(p) ((void *)(p))
+#endif
+
+/* Reasonably portable pointer alignment macro */
+
+#define ALIGNED(P, N) ((void *)((char *)(P) + \
+	((~(unsigned)((char *)(P) - (char *)0) + 1) & ((N) - 1))))
+
+/* Get preferred alignment of a type */
+
+#ifdef __GNUC__ /* Have native implementation */
+#define ALIGNOF(X) __alignof__(X)
+#else
+#define ALIGNOF(X) offsetof(struct {char c; X x;}, x)
 #endif
 
 /* x87 FPU uses long doubles internally, which may cause calculation results
@@ -718,3 +933,6 @@ int average_pixels(unsigned char *rgb, int iw, int ih, int x, int y, int w, int
 #define WJ_ROUND(I,X) (I) = (int)((X) + 0.5)
 #define WJ_FLOOR(I,X) (I) = floor(X)
 #endif
+
+/* Tell GCC to remain silent about an uninitialized variable */
+#define uninit_(X) X = X
