@@ -1,5 +1,5 @@
 /*	otherwindow.h
-	Copyright (C) 2004-2008 Mark Tyler and Dmitry Groshev
+	Copyright (C) 2004-2011 Mark Tyler and Dmitry Groshev
 
 	This file is part of mtPaint.
 
@@ -17,8 +17,6 @@
 	along with mtPaint in the file COPYING.
 */
 
-#include <png.h>
-
 
 #define COLSEL_OVERLAYS  1
 #define COLSEL_EDIT_AB   2
@@ -26,13 +24,18 @@
 #define COLSEL_GRID      4
 #define COLSEL_EDIT_ALL  256
 
+#define CHOOSE_PATTERN 0
+#define CHOOSE_BRUSH   1
+#define CHOOSE_COLOR   2
+
 typedef int (*filter_hook)(GtkWidget *content, gpointer user_data);
 typedef void (*colour_hook)(int what);
 
-png_color brcosa_pal[256];
-int mem_preview, brcosa_auto;
+png_color brcosa_palette[256];
+int mem_preview, mem_preview_clip, brcosa_auto;
 int sharper_reduce;
 int spal_mode;
+seg_state *seg_preview;
 
 void generic_new_window(int type);
 
@@ -43,6 +46,7 @@ void pressed_scale_size(int mode);
 
 void pressed_sort_pal();
 void pressed_quantize(int palette);
+void pressed_pick_gradient();
 
 void choose_pattern(int typ);				// Bring up pattern chooser
 
@@ -60,3 +64,5 @@ void gradient_setup(int mode);
 void pressed_skew();
 
 void bkg_setup();
+
+void pressed_segment();
