@@ -1,5 +1,5 @@
 /*	prefs.c
-	Copyright (C) 2005-2009 Mark Tyler and Dmitry Groshev
+	Copyright (C) 2005-2011 Mark Tyler and Dmitry Groshev
 
 	This file is part of mtPaint.
 
@@ -25,6 +25,7 @@
 #include "inifile.h"
 #include "viewer.h"
 #include "mainwindow.h"
+#include "thread.h"
 
 #include "prefs.h"
 
@@ -33,18 +34,24 @@
 
 GtkWidget *prefs_window, *prefs_status[STATUS_ITEMS];
 static GtkWidget *spinbutton_maxmem, *spinbutton_maxundo, *spinbutton_commundo;
+#ifdef U_THREADS
+static GtkWidget *spinbutton_threads;
+#endif
 static GtkWidget *spinbutton_greys, *spinbutton_nudge, *spinbutton_pan;
 static GtkWidget *spinbutton_trans, *spinbutton_hotx, *spinbutton_hoty,
 	*spinbutton_jpeg, *spinbutton_jp2, *spinbutton_png, *spinbutton_recent,
 	*spinbutton_silence;
 static GtkWidget *checkbutton_tgaRLE, *checkbutton_tga565, *checkbutton_tgadef,
 	*checkbutton_undo;
+#ifdef U_LCMS
+static GtkWidget *checkbutton_icc;
+#endif
 
 static GtkWidget *checkbutton_paste, *checkbutton_cursor, *checkbutton_exit, *checkbutton_quit;
 static GtkWidget *checkbutton_zoom[4],		// zoom 100%, wheel, optimize cheq, disable trans
 	*checkbutton_commit, *checkbutton_center, *checkbutton_gamma;
 #if GTK_MAJOR_VERSION == 2
-static GtkWidget *checkbutton_menuicons;
+static GtkWidget *checkbutton_menuicons, *entry_theme;
 #endif
 static GtkWidget *clipboard_entry, *entry_handbook[2], *entry_def[2];
 static GtkWidget *check_tablet[3], *hscale_tablet[3], *label_tablet_device, *label_tablet_pressure;
@@ -58,6 +65,7 @@ static GdkDeviceInfo *tablet_device;
 #if GTK_MAJOR_VERSION == 2
 static GdkDevice *tablet_device;
 #endif
+static GtkWidget *inputd;
 
 int tablet_working;		// Has the device been initialized?
 
@@ -67,15 +75,15 @@ float tablet_tool_factor[3];			// Size, flow, opacity
 
 #ifdef U_NLS
 
-#define PREF_LANGS 19
+#define PREF_LANGS 21
 
 char *pref_lang_ini_code[PREF_LANGS] = { "system",
 	"zh_CN.utf8", "zh_TW.utf8",
 	"cs_CZ", "nl_NL", "en_GB", "fr_FR",
-	"gl_ES", "de_DE", "it_IT", "ja_JP.utf8",
-	"pl_PL", "pt_PT", "pt_BR",
-	"ru_RU", "sk_SK", "es_ES", "sv_SE",
-	"tr_TR" };
+	"gl_ES", "de_DE", "hu_HU", "it_IT",
+	"ja_JP.utf8", "pl_PL", "pt_PT",
+	"pt_BR", "ru_RU", "sk_SK",
+	"es_ES", "sv_SE", "tl_PH", "tr_TR" };
 
 int pref_lang;
 
@@ -91,18 +99,14 @@ static gboolean expose_tablet_preview(GtkWidget *widget, GdkEventExpose *event)
 }
 
 
-static GtkWidget *inputd = NULL;
-
-
-gint delete_inputd( GtkWidget *widget, GdkEvent *event, gpointer data )
+static void delete_inputd()
 {
 	int i, j;
 	char txt[32];
 
 #if GTK_MAJOR_VERSION == 1
 	GdkDeviceInfo *dev = tablet_device;
-#endif
-#if GTK_MAJOR_VERSION == 2
+#else /* #if GTK_MAJOR_VERSION == 2 */
 	GdkDevice *dev = GTK_INPUT_DIALOG (inputd)->current_device;
 #endif
 
@@ -115,8 +119,7 @@ gint delete_inputd( GtkWidget *widget, GdkEvent *event, gpointer data )
 		{
 #if GTK_MAJOR_VERSION == 1
 			j = dev->axes[i];
-#endif
-#if GTK_MAJOR_VERSION == 2
+#else /* #if GTK_MAJOR_VERSION == 2 */
 			j = dev->axes[i].use;
 #endif
 			sprintf(txt, "tablet_axes_v%i", i);
@@ -126,14 +129,12 @@ gint delete_inputd( GtkWidget *widget, GdkEvent *event, gpointer data )
 
 	gtk_widget_destroy(inputd);
 	inputd = NULL;
-
-	return FALSE;
 }
 
 static void delete_prefs(GtkWidget *widget)
 {
-	if ( inputd != NULL ) delete_inputd( NULL, NULL, NULL );
-	gtk_widget_destroy(prefs_window);
+	if (inputd) delete_inputd();
+	destroy_dialog(prefs_window);
 	gtk_widget_set_sensitive(menu_widgets[MENU_PREFS], TRUE);
 	clipboard_entry = NULL;
 }
@@ -208,39 +209,38 @@ static void tablet_disable_device(GtkInputDialog *inputdialog, GdkDevice *device
 #endif
 
 
-gint conf_tablet( GtkWidget *widget, GdkEvent *event, gpointer data )
+static void conf_tablet(GtkWidget *widget)
 {
-	GtkAccelGroup* ag = gtk_accel_group_new();
+	GtkWidget *close;
+	GtkAccelGroup* ag;
 
-	if (inputd != NULL) return FALSE;	// Stops multiple dialogs being opened
+	if (inputd) return;	// Stops multiple dialogs being opened
 
 	inputd = gtk_input_dialog_new();
-	gtk_window_set_position( GTK_WINDOW(inputd), GTK_WIN_POS_CENTER );
+	gtk_window_set_position(GTK_WINDOW(inputd), GTK_WIN_POS_CENTER);
 
-	gtk_signal_connect(GTK_OBJECT (GTK_INPUT_DIALOG (inputd)->close_button), "clicked",
-		GTK_SIGNAL_FUNC(delete_inputd), (gpointer) inputd);
-	gtk_widget_add_accelerator (GTK_INPUT_DIALOG (inputd)->close_button, "clicked",
-		ag, GDK_Escape, 0, (GtkAccelFlags) 0);
+	close = GTK_INPUT_DIALOG(inputd)->close_button;
+	gtk_signal_connect(GTK_OBJECT(close), "clicked",
+		GTK_SIGNAL_FUNC(delete_inputd), NULL);
+	ag = gtk_accel_group_new();
+	gtk_widget_add_accelerator(close, "clicked",
+		ag, GDK_Escape, 0, (GtkAccelFlags)0);
+	delete_to_click(inputd, close);
 
-	gtk_signal_connect(GTK_OBJECT (inputd), "destroy",
-		GTK_SIGNAL_FUNC(delete_inputd), (gpointer) inputd);
+	gtk_signal_connect(GTK_OBJECT(inputd), "enable-device",
+		GTK_SIGNAL_FUNC(tablet_enable_device), NULL);
+	gtk_signal_connect(GTK_OBJECT(inputd), "disable-device",
+		GTK_SIGNAL_FUNC(tablet_disable_device), NULL);
 
-	gtk_signal_connect(GTK_OBJECT (inputd), "enable-device",
-		GTK_SIGNAL_FUNC(tablet_enable_device), (gpointer) inputd);
-	gtk_signal_connect(GTK_OBJECT (inputd), "disable-device",
-		GTK_SIGNAL_FUNC(tablet_disable_device), (gpointer) inputd);
+	if (GTK_INPUT_DIALOG(inputd)->keys_list)
+		gtk_widget_hide(GTK_INPUT_DIALOG(inputd)->keys_list);
+	if (GTK_INPUT_DIALOG(inputd)->keys_listbox)
+		gtk_widget_hide(GTK_INPUT_DIALOG(inputd)->keys_listbox);
 
-	if ( GTK_INPUT_DIALOG (inputd)->keys_list != NULL )
-		gtk_widget_hide (GTK_INPUT_DIALOG (inputd)->keys_list);
-	if ( GTK_INPUT_DIALOG (inputd)->keys_listbox != NULL )
-		gtk_widget_hide (GTK_INPUT_DIALOG (inputd)->keys_listbox);
+	gtk_widget_hide(GTK_INPUT_DIALOG(inputd)->save_button);
 
-	gtk_widget_hide (GTK_INPUT_DIALOG (inputd)->save_button);
-
-	gtk_widget_show (inputd);
-	gtk_window_add_accel_group(GTK_WINDOW (inputd), ag);
-
-	return FALSE;
+	gtk_widget_show(inputd);
+	gtk_window_add_accel_group(GTK_WINDOW(inputd), ag);
 }
 
 
@@ -248,7 +248,7 @@ gint conf_tablet( GtkWidget *widget, GdkEvent *event, gpointer data )
 static void prefs_apply(GtkWidget *widget)
 {
 	char path[PATHBUF];
-	int i, j;
+	int i, j, xpm_trans;
 
 	for ( i=0; i<STATUS_ITEMS; i++ )
 	{
@@ -261,8 +261,10 @@ static void prefs_apply(GtkWidget *widget)
 	mem_background = read_spin(spinbutton_greys);
 	mem_nudge = read_spin(spinbutton_nudge);
 	max_pan = read_spin(spinbutton_pan);
-
-	mem_xpm_trans = read_spin(spinbutton_trans);
+#ifdef U_THREADS
+	maxthreads = read_spin(spinbutton_threads);
+#endif
+	xpm_trans = read_spin(spinbutton_trans);
 	mem_xbm_hot_x = read_spin(spinbutton_hotx);
 	mem_xbm_hot_y = read_spin(spinbutton_hoty);
 	jpeg_quality = read_spin(spinbutton_jpeg);
@@ -285,6 +287,9 @@ static void prefs_apply(GtkWidget *widget)
 	tga_565 = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_tga565));
 	tga_defdir = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_tgadef));
 	undo_load = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_undo));
+#ifdef U_LCMS
+	apply_icc = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_icc));
+#endif
 
 	show_paste = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_paste));
 	cursor_tool = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_cursor));
@@ -306,8 +311,7 @@ static void prefs_apply(GtkWidget *widget)
 
 	inifile_set_gboolean("centerSettings",
 		gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_center)));
-	inifile_set_gboolean("defaultGamma",
-		gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_gamma)));
+	use_gamma = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_gamma));
 #if GTK_MAJOR_VERSION == 2
 	show_menu_icons = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_menuicons));
 #endif
@@ -338,7 +342,14 @@ static void prefs_apply(GtkWidget *widget)
 		load_def_patterns(path);
 	inifile_set(DEFAULT_PAT_INI, path);
 
+#if GTK_MAJOR_VERSION == 2
+	gtkncpy(path, gtk_entry_get_text(GTK_ENTRY(entry_theme)), PATHBUF);
+	inifile_set(DEFAULT_THEME_INI, path);
+#endif
+
 	update_stuff(UPD_PREFS);
+	/* Apply this undoable setting after everything else */
+	mem_set_trans(xpm_trans);
 }
 
 static void prefs_ok(GtkWidget *widget)
@@ -365,7 +376,7 @@ static gint tablet_preview_button (GtkWidget *widget, GdkEventButton *event)
 	return TRUE;
 }
 
-static gint tablet_preview_motion(GtkWidget *widget, GdkEventMotion *event)
+static gboolean tablet_preview_motion(GtkWidget *widget, GdkEventMotion *event)
 {
 	gdouble pressure = 0.0;
 	GdkModifierType state;
@@ -381,8 +392,7 @@ static gint tablet_preview_motion(GtkWidget *widget, GdkEventMotion *event)
 		pressure = event->pressure;
 		state = event->state;
 	}
-#endif
-#if GTK_MAJOR_VERSION == 2
+#else /* #if GTK_MAJOR_VERSION == 2 */
 	if (event->is_hint) gdk_device_get_state (event->device, event->window, NULL, &state);
 	else state = event->state;
 
@@ -395,6 +405,25 @@ static gint tablet_preview_motion(GtkWidget *widget, GdkEventMotion *event)
 	return TRUE;
 }
 
+/* Try to avoid scrolling - request full size of contents */
+static void pref_scroll_size_req(GtkWidget *widget, GtkRequisition *requisition,
+	gpointer user_data)
+{
+	GtkWidget *child = GTK_BIN(widget)->child;
+
+	if (child && GTK_WIDGET_VISIBLE(child))
+	{
+		GtkRequisition wreq;
+		int n, border = GTK_CONTAINER(widget)->border_width * 2;
+
+		gtk_widget_get_child_requisition(child, &wreq);
+		n = wreq.width + border;
+		if (requisition->width < n) requisition->width = n;
+		n = wreq.height + border;
+		if (requisition->height < n) requisition->height = n;
+	}
+}
+
 void pressed_preferences()
 {
 	int i;
@@ -402,15 +431,15 @@ void pressed_preferences()
 	char *pref_langs[PREF_LANGS] = { _("Default System Language"),
 		_("Chinese (Simplified)"), _("Chinese (Taiwanese)"),
 		_("Czech"), _("Dutch"), _("English (UK)"), _("French"),
-		_("Galician"), _("German"), _("Italian"), _("Japanese"),
-		_("Polish"), _("Portuguese"), _("Portuguese (Brazilian)"),
-		_("Russian"), _("Slovak"), _("Spanish"), _("Swedish"),
-		_("Turkish") };
+		_("Galician"), _("German"), _("Hungarian"), _("Italian"),
+		_("Japanese"), _("Polish"), _("Portuguese"),
+		_("Portuguese (Brazilian)"), _("Russian"), _("Slovak"),
+		_("Spanish"), _("Swedish"), _("Tagalog"), _("Turkish") };
 #endif
 
 
 	GtkWidget *vbox3, *hbox4, *table3, *table4, *drawingarea_tablet;
-	GtkWidget *button1, *notebook1, *page, *vbox_2, *label;
+	GtkWidget *button1, *notebook1, *page, *vbox_2, *label, *scroll;
 
 	char *tab_tex2[] = { _("Transparency index"), _("XBM X hotspot"), _("XBM Y hotspot"),
 		_("JPEG Save Quality (100=High)"), _("JPEG2000 Compression (0=Lossless)"),
@@ -426,22 +455,38 @@ void pressed_preferences()
 	gtk_widget_set_sensitive(menu_widgets[MENU_PREFS], FALSE);
 
 	prefs_window = add_a_window( GTK_WINDOW_TOPLEVEL, _("Preferences"), GTK_WIN_POS_CENTER, FALSE );
+	// Let user and WM make window smaller - contents are scrollable
+	gtk_window_set_policy(GTK_WINDOW(prefs_window), TRUE, TRUE, FALSE);
+	vbox3 = add_vbox(prefs_window);
 
-	vbox3 = gtk_vbox_new (FALSE, 0);
-	gtk_widget_show (vbox3);
-	gtk_container_add (GTK_CONTAINER (prefs_window), vbox3);
+///	SETUP SCROLLING
+
+	scroll = xpack(vbox3, gtk_scrolled_window_new(NULL, NULL));
+	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
+		GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
+	gtk_signal_connect(GTK_OBJECT(scroll), "size_request",
+		GTK_SIGNAL_FUNC(pref_scroll_size_req), NULL);
 
 ///	SETUP NOTEBOOK
-
-	notebook1 = xpack(vbox3, gtk_notebook_new());
-	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(notebook1), GTK_POS_TOP);
+	
+	notebook1 = gtk_notebook_new();
+	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(notebook1), GTK_POS_LEFT);
 //	gtk_notebook_set_scrollable(GTK_NOTEBOOK(notebook1), TRUE);
-	gtk_widget_show(notebook1);
+	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scroll), notebook1);
+	gtk_viewport_set_shadow_type(GTK_VIEWPORT(GTK_BIN(scroll)->child), GTK_SHADOW_NONE);
+	gtk_widget_show_all(scroll);
+	vport_noshadow_fix(GTK_BIN(scroll)->child);
 
 ///	---- TAB1 - GENERAL
 
 	page = add_new_page(notebook1, _("General"));
+#ifdef U_THREADS
+	table3 = add_a_table(4, 2, 10, page);
+	add_to_table(_("Max threads (0 to autodetect)"), table3, 3, 0, 5);
+	spinbutton_threads = spin_to_table(table3, 3, 1, 5, maxthreads, 0, 256);
+#else
 	table3 = add_a_table(3, 2, 10, page);
+#endif
 
 ///	TABLE TEXT
 	add_to_table(_("Max memory used for undo (MB)"), table3, 0, 0, 5);
@@ -455,7 +500,7 @@ void pressed_preferences()
 	spinbutton_commundo = spin_to_table(table3, 2, 1, 5, mem_undo_common, 0, 100);
 
 	checkbutton_gamma = add_a_toggle(_("Use gamma correction by default"),
-		page, inifile_get_gboolean("defaultGamma", FALSE));
+		page, use_gamma);
 	checkbutton_zoom[2] = add_a_toggle( _("Optimize alpha chequers"),
 		page, chequers_optimize );
 	checkbutton_zoom[3] = add_a_toggle( _("Disable view window transparencies"),
@@ -477,7 +522,7 @@ void pressed_preferences()
 	pack(vbox_2, wj_option_menu(pref_langs, PREF_LANGS, i, &pref_lang, NULL));
 
 	gtk_widget_show_all(vbox_2);
-	add_with_frame(page, _("Language"), vbox_2, 5);
+	add_with_frame(page, _("Language"), vbox_2);
 #endif
 
 ///	---- TAB2 - INTERFACE
@@ -535,6 +580,9 @@ void pressed_preferences()
 	checkbutton_tga565 = add_a_toggle(_("Read 16-bit TGAs as 5:6:5 BGR"), page, tga_565);
 	checkbutton_tgadef = add_a_toggle(_("Write TGAs in bottom-up row order"), page, tga_defdir);
 	checkbutton_undo   = add_a_toggle(_("Undoable image loading"), page, undo_load);
+#ifdef U_LCMS
+	checkbutton_icc    = add_a_toggle(_("Apply colour profile"), page, apply_icc);
+#endif
 
 ///	---- TAB4 - PATHS
 
@@ -565,6 +613,12 @@ void pressed_preferences()
 	gtkuncpy(txt, inifile_get(DEFAULT_PAT_INI, ""), PATHTXT);
 	gtk_entry_set_text(GTK_ENTRY(entry_def[1]), txt);
 
+#if GTK_MAJOR_VERSION == 2
+	entry_theme = mt_path_box(_("Default Theme"), page,
+		_("Select Default Theme File"), FS_SELECT_FILE);
+	gtkuncpy(txt, inifile_get(DEFAULT_THEME_INI, ""), PATHTXT);
+	gtk_entry_set_text(GTK_ENTRY(entry_theme), txt);
+#endif
 
 ///	---- TAB5 - STATUS BAR
 
@@ -581,7 +635,7 @@ void pressed_preferences()
 
 	vbox_2 = gtk_vbox_new (FALSE, 0);
 	gtk_widget_show (vbox_2);
-	add_with_frame(page, _("Device Settings"), vbox_2, 5);
+	add_with_frame(page, _("Device Settings"), vbox_2);
 	gtk_container_set_border_width (GTK_CONTAINER (vbox_2), 5);
 
 	label_tablet_device = pack(vbox_2, gtk_label_new(""));
@@ -590,7 +644,8 @@ void pressed_preferences()
 	gtk_misc_set_padding (GTK_MISC (label_tablet_device), 5, 5);
 
 	button1 = add_a_button( _("Configure Device"), 0, vbox_2, FALSE );
-	gtk_signal_connect(GTK_OBJECT(button1), "clicked", GTK_SIGNAL_FUNC(conf_tablet), NULL);
+	gtk_signal_connect(GTK_OBJECT(button1), "clicked",
+		GTK_SIGNAL_FUNC(conf_tablet), NULL);
 
 	table3 = xpack(vbox_2, gtk_table_new(4, 2, FALSE));
 	gtk_widget_show (table3);
@@ -604,13 +659,11 @@ void pressed_preferences()
 
 	for ( i=0; i<3; i++ )
 	{
-		check_tablet[i] = gtk_check_button_new_with_label (tablet_txt[i]);
+		check_tablet[i] = gtk_check_button_new_with_label(tablet_txt[i]);
 		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check_tablet[i]),
-			tablet_tool_use[i] );
-		gtk_widget_show (check_tablet[i]);
-		gtk_table_attach (GTK_TABLE (table3), check_tablet[i], 0, 1, i+1, i+2,
-			(GtkAttachOptions) (GTK_FILL),
-			(GtkAttachOptions) (0), 0, 0);
+			tablet_tool_use[i]);
+		gtk_widget_show(check_tablet[i]);
+		to_table_l(check_tablet[i], table3, i + 1, 0, 1, 0);
 
 //	Size/Flow/Opacity sliders
 
@@ -625,7 +678,7 @@ void pressed_preferences()
 
 	vbox_2 = gtk_vbox_new (FALSE, 0);
 	gtk_widget_show (vbox_2);
-	add_with_frame(page, _("Test Area"), vbox_2, 5);
+	add_with_frame(page, _("Test Area"), vbox_2);
 	gtk_container_set_border_width (GTK_CONTAINER (vbox_2), 5);
 
 	drawingarea_tablet = xpack(vbox_2, gtk_drawing_area_new());
@@ -659,15 +712,12 @@ void pressed_preferences()
 
 	hbox4 = pack(vbox3, OK_box(0, prefs_window, _("OK"), GTK_SIGNAL_FUNC(prefs_ok),
 		_("Cancel"), GTK_SIGNAL_FUNC(delete_prefs)));
-	OK_box_add(hbox4, _("Apply"), GTK_SIGNAL_FUNC(prefs_apply), 1);
+	OK_box_add(hbox4, _("Apply"), GTK_SIGNAL_FUNC(prefs_apply));
 
 	gtk_window_set_transient_for( GTK_WINDOW(prefs_window), GTK_WINDOW(main_window) );
 	gtk_widget_show (prefs_window);
 
-	if ( tablet_working )
-	{
-		tablet_update_device( tablet_device->name );
-	} else	tablet_update_device( "NONE" );
+	tablet_update_device(tablet_working ? tablet_device->name : "NONE");
 }
 
 
