@@ -1,5 +1,5 @@
 /*	spawn.c
-	Copyright (C) 2007-2009 Mark Tyler and Dmitry Groshev
+	Copyright (C) 2007-2011 Mark Tyler and Dmitry Groshev
 
 	This file is part of mtPaint.
 
@@ -17,9 +17,7 @@
 	along with mtPaint in the file COPYING.
 */
 
-#include <unistd.h>
 #include <fcntl.h>
-#include <sys/stat.h>
 
 #include "global.h"
 
@@ -38,47 +36,49 @@ static char *get_tempdir()
 	char *env;
 
 	env = getenv("TMPDIR");
-	if (!env) env = getenv("TMP");
-	if (!env) env = getenv("TEMP");
+	if (!env || !*env) env = getenv("TMP");
+	if (!env || !*env) env = getenv("TEMP");
 #ifdef P_tmpdir
-	if (!env) env = P_tmpdir;
+	if (!env || !*env) env = P_tmpdir;
 #endif
+	if (!env || !*env || (strlen(env) >= PATHBUF)) // Bad if too long
 #ifdef WIN32
-	if (!env) env = "\\";
+		env = "\\";
 #else
-	if (!env) env = "/tmp";
+		env = "/tmp";
 #endif
 	return (env);
 }
 
 static char *new_temp_dir()
 {
-	char buf[PATHBUF], *base = get_tempdir();
-	int l = strlen(base);
+	char *buf, *base = get_tempdir();
 
-	l -= base[l - 1] == DIR_SEP;
 #ifdef HAVE_MKDTEMP
-	snprintf(buf, PATHBUF, "%.*s%cmtpaintXXXXXX", l, base, DIR_SEP);
-	if (!mkdtemp(buf)) return (NULL);
-	chmod(buf, 0755);
-	return (strdup(buf));
+	buf = file_in_dir(NULL, base, "mtpaintXXXXXX", PATHBUF);
+	if (!buf) return (NULL);
+	if (mkdtemp(buf))
+	{
+		chmod(buf, 0755);
+		return (buf);
+	}
 #else
-	strncpy0(buf, base, l); /* Cut off path separator */
-	base = tempnam(buf, "mttmp");
-	if (!base) return (NULL);
+	buf = tempnam(base, "mttmp");
+	if (!buf) return (NULL);
 #ifdef WIN32 /* No mkdtemp() in MinGW */
 	/* tempnam() may use Unix path separator */
-	for (l = 0; base[l]; l++) if (base[l] == '/') base[l] = '\\';
-	if (mkdir(base)) return (NULL);
+	reseparate(buf);
+	if (!mkdir(buf)) return (buf);
 #else
-	if (mkdir(base, 0755)) return (NULL);
+	if (!mkdir(buf, 0755)) return (buf);
 #endif
-	return (base);
 #endif
+	free(buf);
+	return (NULL);
 }
 
 /* Store index for name, or fetch it (when idx < 0) */
-static int last_temp_index(char *name, int len, int idx)
+static int last_temp_index(char *name, int idx)
 {
 // !!! For now, use simplest model - same index regardless of name
 	static int index;
@@ -86,64 +86,111 @@ static int last_temp_index(char *name, int len, int idx)
 	return (index);
 }
 
-/* I'm lazy today, so using off-the-shelf list - WJ */
-static GList *namelist;
+typedef struct {
+	void *next, *id;
+	int type, rgb;
+	char name[1];
+} tempfile;
 
-static char *remember_temp_file(char *name)
+static void *tempchain;
+
+static char *remember_temp_file(char *name, int type, int rgb)
 {
-	namelist = g_list_prepend(namelist, name = strdup(name));
-	return (name);
+	static wjmem *tempstore;
+	tempfile *tmp, *tm0;
+
+	if ((tempstore || (tempstore = wjmemnew(0, 0))) &&
+		(tmp = wjmalloc(tempstore, offsetof(tempfile, name) +
+		strlen(name) + 1, ALIGNOF(tempfile))))
+	{
+		tmp->type = type;
+		tmp->rgb = rgb;
+		strcpy(tmp->name, name);
+		if (mem_tempfiles) /* Have anchor - insert after it */
+		{
+			tm0 = tmp->id = mem_tempfiles;
+			tmp->next = tm0->next;
+			tm0->next = (void *)tmp;
+		}
+		else /* No anchor - insert before all */
+		{
+			tmp->next = tempchain;
+			mem_tempfiles = tempchain = tmp->id = (void *)tmp;
+		}
+		return (tmp->name);
+	}
+	return (NULL); // Failed to allocate
 }
 
 void spawn_quit()
 {
-	g_list_foreach(namelist, (GFunc)unlink, NULL);
+	tempfile *tmp;
+
+	for (tmp = tempchain; tmp; tmp = tmp->next) unlink(tmp->name);
 	if (mt_temp_dir) rmdir(mt_temp_dir);
 }
 
-static char *new_temp_file()
+static char *get_temp_file(int type, int rgb)
 {
 	ls_settings settings;
-	char buf[PATHBUF], ids[32], *c, *f = "tmp.png", *ext = "png";
-	int fd, l, cnt, idx, res, type = FT_PNG;
+	tempfile *tmp;
+	unsigned char *img = NULL;
+	char buf[PATHBUF], nstub[NAMEBUF], ids[32], *c, *f = "tmp.png";
+	int fd, cnt, idx, res;
+
+	/* Use the original file if possible */
+	if (!mem_changed && mem_filename && (!rgb ^ (mem_img_bpp == 3)) &&
+		((type == FT_NONE) ||
+		 (detect_file_format(mem_filename, FALSE) == type)))
+		return (mem_filename);
 
 	/* Prepare temp directory */
 	if (!mt_temp_dir) mt_temp_dir = new_temp_dir();
 	if (!mt_temp_dir) return (NULL); /* Temp dir creation failed */
 
 	/* Analyze name */
-	if (mem_filename[0])
+	if ((type == FT_NONE) && mem_filename)
 	{
 		f = strrchr(mem_filename, DIR_SEP);
 		if (!f) f = mem_filename;
 		type = file_type_by_ext(f, FF_SAVE_MASK);
-		if (type == FT_NONE) type = FT_PNG;
-		else ext = strrchr(f, '.') + 1;
 	}
+	if (type == FT_NONE) type = FT_PNG;
+
+	/* Use existing file if possible */
+	for (tmp = mem_tempfiles; tmp; tmp = tmp->next)
+	{
+		if (tmp->id != mem_tempfiles) break;
+		if ((tmp->type == type) && (tmp->rgb == rgb))
+			return (tmp->name);
+	}
+
+	/* Stubify filename */
+	strcpy(nstub, "tmp");
 	c = strrchr(f, '.');
-	if (c == f) c = (f = "tmp.png") + 3; /* If extension w/o name */
-	l = c ? c - f : strlen(f);
+	if (c != f) /* Extension should not be alone */
+		wjstrcat(nstub, NAMEBUF, f, c ? c - f : strlen(f), NULL);
 
 	/* Create temp file */
 	while (TRUE)
 	{
-		idx = last_temp_index(f, l, -1);
+		idx = last_temp_index(nstub, -1);
 		ids[0] = 0;
 		for (cnt = 0; cnt < 256; cnt++ , idx++)
 		{
 			if (idx) sprintf(ids, "%d", idx);
-			snprintf(buf, PATHBUF, "%s%c%.*s%s.%s",
-				mt_temp_dir, DIR_SEP, l, f, ids, ext);
+			snprintf(buf, PATHBUF, "%s" DIR_SEP_STR "%s%s.%s",
+				mt_temp_dir, nstub, ids, file_formats[type].ext);
 			fd = open(buf, O_WRONLY | O_CREAT | O_EXCL, 0644);
 			if (fd >= 0) break;
 		}
-		last_temp_index(f, l, idx);
+		last_temp_index(nstub, idx);
 		if (fd >= 0) break;
-		if (!strncmp(f, "tmp.png", l)) return (NULL); /* Utter failure */
-		f = "tmp.png"; l = 3; /* Try again with "tmp" */
+		if (!strcmp(nstub, "tmp")) return (NULL); /* Utter failure */
+		strcpy(nstub, "tmp"); /* Try again with "tmp" */
 	}
 	close(fd);
-	
+
 	/* Save image */
 	init_ls_settings(&settings, NULL);
 	memcpy(settings.img, mem_img, sizeof(chanlist));
@@ -153,25 +200,67 @@ static char *new_temp_file()
 	settings.bpp = mem_img_bpp;
 	settings.colors = mem_cols;
 	settings.ftype = type;
+	if (rgb && (mem_img_bpp == 1)) /* Save indexed as RGB */
+	{
+		settings.img[CHN_IMAGE] = img =
+			malloc(mem_width * mem_height * 3);
+		if (!img) return (NULL); /* Failed to allocate RGB buffer */
+		settings.bpp = 3;
+		do_convert_rgb(0, 1, mem_width * mem_height, img,
+			mem_img[CHN_IMAGE], mem_pal);
+	}
 	res = save_image(buf, &settings);
+	free(img);
 	if (res) return (NULL); /* Failed to save */
 
-	return (remember_temp_file(buf));
+	return (remember_temp_file(buf, type, rgb));
 }
 
 static char *insert_temp_file(char *pattern, int where, int skip)
 {
-	char *fname = mem_filename;
+	char *fname, *pat = pattern;
+	int i, l, rgb = mem_img_bpp == 3, fform = FT_NONE;
 
-	if (mem_changed || !fname[0]) /* If not saved */
+	while (TRUE)
 	{
-		if (!mem_tempname) mem_tempname = new_temp_file();
-		if (!mem_tempname) return (NULL); /* Temp save failed */
-		fname = mem_tempname;
+		/* Skip initial whitespace if any */
+		pat += strspn(pat, " \t");
+		/* Finish if not a transform request */
+		if (*pat != '>') break;
+		l = strcspn(++pat, "> \t");
+		if (!strncasecmp("RGB", pat, l)) rgb = TRUE;
+		else
+		{
+			for (i = FT_NONE + 1; i < NUM_FTYPES; i++)
+			{
+				fformat *ff = file_formats + i;
+				if ((ff->flags & FF_IMAGE) &&
+					!(ff->flags & FF_NOSAVE) &&
+					(!strncasecmp(ff->name, pat, l) ||
+					!strncasecmp(ff->ext, pat, l) ||
+					(ff->ext2[0] &&
+					!strncasecmp(ff->ext2, pat, l))))
+					fform = i;
+			}
+		}
+		pat += l;
+	}
+	where -= pat - pattern;
+	if (where < 0) return (NULL); // Syntax error
+
+	if (fform != FT_NONE)
+	{
+		unsigned int flags = file_formats[fform].flags;
+		if (rgb && !(flags & FF_RGB)) fform = FT_NONE; // No way
+		else if (flags & FF_SAVE_MASK); // Is OK
+		else if (flags & FF_RGB) rgb = TRUE; // Fallback
+		else fform = FT_NONE; // Give up
 	}
 
-	return (g_strdup_printf("%.*s\"%s\"%s", where, pattern, fname,
-		pattern + where + skip));
+	fname = get_temp_file(fform, rgb);
+	if (!fname) return (NULL); /* Temp save failed */
+	return (wjstrcat(NULL, 0, pat, where, "\"", fname, "\"",
+		pat + where + skip, NULL));
 }
 
 int spawn_expansion(char *cline, char *directory)
@@ -193,7 +282,7 @@ int spawn_expansion(char *cline, char *directory)
 		{
 			argv[2] = s1;
 			res = spawn_process(argv, directory);
-			g_free(s1);
+			free(s1);
 		}
 		else return -1;
 	}
@@ -376,8 +465,7 @@ void pressed_file_configure()
 	win = add_a_window( GTK_WINDOW_TOPLEVEL, _("Configure File Actions"), GTK_WIN_POS_CENTER, TRUE );
 	gtk_window_set_default_size( GTK_WINDOW(win), 500, 400 );
 
-	vbox = gtk_vbox_new(FALSE, 0);
-	gtk_container_add(GTK_CONTAINER(win), vbox);
+	vbox = add_vbox(win);
 	gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
 
 	sw = xpack5(vbox, gtk_scrolled_window_new(NULL, NULL));
@@ -418,7 +506,7 @@ void pressed_file_configure()
 	{
 		hbox = gtk_hbox_new(FALSE, 0);
 		gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);
-		add_with_frame(vbox, !j ? _("Action") : _("Command"), hbox, 5);
+		add_with_frame(vbox, !j ? _("Action") : _("Command"), hbox);
 		faction_entry[j] = entry = xpack5(hbox, gtk_entry_new());
 		gtk_signal_connect(GTK_OBJECT(entry), "changed",
 			GTK_SIGNAL_FUNC(faction_changed), (gpointer)j);
@@ -433,7 +521,7 @@ void pressed_file_configure()
 
 	hbox = pack(vbox, OK_box(0, win, _("OK"), GTK_SIGNAL_FUNC(faction_ok),
 		_("Cancel"), GTK_SIGNAL_FUNC(gtk_widget_destroy)));
-	OK_box_add(hbox, _("Execute"), GTK_SIGNAL_FUNC(faction_execute), 1);
+	OK_box_add(hbox, _("Execute"), GTK_SIGNAL_FUNC(faction_execute));
 
 	gtk_window_set_transient_for(GTK_WINDOW(win), GTK_WINDOW(main_window));
 	gtk_widget_show_all(win);
@@ -544,6 +632,46 @@ int spawn_process(char *argv[], char *directory)
 
 #endif
 
+// Copy string quoting space chars
+
+static char *quote_spaces(const char *src)
+{
+	char *tmp, *dest;
+	int n = 0;
+
+	for (*(const char **)&tmp = src; *tmp; tmp++)
+	{
+		if (*tmp == ' ')
+#ifdef WIN32
+			n += 2;
+#else
+			n++;
+#endif
+		n++;
+	}
+
+	dest = malloc(n + 1);
+	if (!dest) return ("");
+
+	for (tmp = dest; *src; src++)
+	{
+		if (*src == ' ')
+		{
+#ifdef WIN32
+			*tmp++ = '"';
+			*tmp++ = ' ';
+			*tmp++ = '"';
+#else
+			*tmp++ = '\\';
+			*tmp++ = ' ';
+#endif
+		}
+		else *tmp++ = *src;
+	}
+	*tmp = 0;
+	return (dest);
+}
+
 // Wrapper for animated GIF handling, or other builtin actions
 
 #ifdef U_ANIM_IMAGICK
@@ -552,8 +680,6 @@ int spawn_process(char *argv[], char *directory)
 	/* !!! Image Magick version prior to 6.2.6-3 won't work properly */
 #define CMD_GIF_CREATE \
 	"convert %2$s -layers optimize -set delay %1$d -loop 0 \"%3$s\""
-#define CMD_GIF_EXPLODE \
-	"convert \"%s\" -scene 0 -coalesce +adjoin \"gif:%s.%%03d\""
 #define CMD_GIF_PLAY "animate \"%s\" &"
 
 #else	/* Use Gifsicle; the default choice for animated GIFs */
@@ -562,16 +688,22 @@ int spawn_process(char *argv[], char *directory)
  *	removal method, infinite loops, ensure result works with Java & MS IE */
 #define CMD_GIF_CREATE \
 	"gifsicle --colors 256 -w -O2 -D 2 -l0 --careful -d %d %s -o \"%s\""
-#define CMD_GIF_EXPLODE \
-	"gifsicle -U --explode \"%s\" -o \"%s\""
 #define CMD_GIF_PLAY "gifview -a \"%s\" &"
 
 #endif
 
-#ifdef WIN32
+/* Windows shell does not know "&", nor needs it to run mtPaint detached */
+#ifndef WIN32
+
+#define CMD_DETACH " &"
+
+#else /* WIN32 */
+
+#define CMD_DETACH
 #ifndef WEXITSTATUS
 #define WEXITSTATUS(A) ((A) & 0xFF)
 #endif
+
 #endif
 
 int run_def_action(int action, char *sname, char *dname, int delay)
@@ -582,10 +714,9 @@ int run_def_action(int action, char *sname, char *dname, int delay)
 	switch (action)
 	{
 	case DA_GIF_CREATE:
-		command = g_strdup_printf(CMD_GIF_CREATE, delay, sname, dname);
-		break;
-	case DA_GIF_EXPLODE:
-		command = g_strdup_printf(CMD_GIF_EXPLODE, sname, dname);
+		c8 = quote_spaces(sname);
+		command = g_strdup_printf(CMD_GIF_CREATE, delay, c8, dname);
+		free(c8);
 		break;
 	case DA_GIF_PLAY:
 #if (GTK_MAJOR_VERSION == 1) || defined GDK_WINDOWING_X11
@@ -596,11 +727,8 @@ int run_def_action(int action, char *sname, char *dname, int delay)
 		return (-1);
 #endif
 	case DA_GIF_EDIT:
-#ifndef WIN32
-		command = g_strdup_printf("mtpaint -g %d %s &", delay, sname);
-#else /* Windows shell does not know "&", nor needs it to run mtPaint detached */
-		command = g_strdup_printf("mtpaint -g %d %s", delay, sname);
-#endif
+		command = g_strdup_printf("mtpaint -g %d -w \"%s.???\" -w \"%s.????\""
+			CMD_DETACH, delay, sname, sname);
 		break;
 	}
 
@@ -612,7 +740,7 @@ int run_def_action(int action, char *sname, char *dname, int delay)
 		c8 = gtkuncpy(NULL, command, 0);
 		msg = g_strdup_printf(_("Error %i reported when trying to run %s"),
 			code, c8);
-		alert_box(_("Error"), msg, _("OK"), NULL, NULL);
+		alert_box(_("Error"), msg, NULL);
 		g_free(msg);
 		g_free(c8);
 	}
@@ -620,3 +748,92 @@ int run_def_action(int action, char *sname, char *dname, int delay)
 
 	return (res);
 }
+
+
+#ifdef WIN32
+
+#include <shellapi.h>
+#define HANDBOOK_LOCATION_WIN "..\\docs\\index.html"
+
+#else /* Linux */
+
+#define HANDBOOK_BROWSER "firefox"
+#define HANDBOOK_LOCATION "/usr/doc/mtpaint/index.html"
+#define HANDBOOK_LOCATION2 "/usr/share/doc/mtpaint/index.html"
+
+#endif
+
+int show_html(char *browser, char *docs)
+{
+	char buf[PATHBUF + 2], buf2[PATHBUF];
+	int i=-1;
+#ifdef WIN32
+	char *r;
+
+	if (!docs || !docs[0])
+	{
+		/* Use default path relative to installdir */
+		docs = buf + 1;
+		i = GetModuleFileNameA(NULL, docs, PATHBUF);
+		if (!i) return (-1);
+		r = strrchr(docs, '\\');
+		if (!r) return (-1);
+		r[1] = 0;
+		strnncat(docs, HANDBOOK_LOCATION_WIN, PATHBUF);
+	}
+#else /* Linux */
+	char *argv[5];
+
+	if (!docs || !docs[0])
+	{
+		docs = HANDBOOK_LOCATION;
+		if (valid_file(docs) < 0) docs = HANDBOOK_LOCATION2;
+	}
+#endif
+	else docs = gtkncpy(buf + 1, docs, PATHBUF);
+
+	if ((valid_file(docs) < 0))
+	{
+		alert_box( _("Error"),
+			_("I am unable to find the documentation.  Either you need to download the mtPaint Handbook from the web site and install it, or you need to set the correct location in the Preferences window."), NULL);
+		return (-1);
+	}
+
+#ifdef WIN32
+	if (browser && !browser[0]) browser = NULL;
+	if (browser)
+	{
+		/* Quote the filename */
+		i = strlen(docs);
+		buf[0] = docs[i] = '"';
+		docs[i + 1] = '\0';
+		/* Rearrange parameters */
+		docs = gtkncpy(buf2, browser, PATHBUF);
+		browser = buf;
+	}
+	if ((unsigned int)ShellExecuteA(NULL, "open", docs, browser,
+		NULL, SW_SHOW) <= 32) i = -1;
+	else i = 0;
+#else
+	argv[1] = docs;
+	argv[2] = NULL;
+	/* Try to use default browser */
+	if (!browser || !browser[0])
+	{
+		argv[0] = "xdg-open";
+		i = spawn_process(argv, NULL);
+		if (!i) return (0); // System has xdg-utils installed
+		// No xdg-utils - try "BROWSER" variable then
+		browser = getenv("BROWSER");
+	}
+	else browser = gtkncpy(buf2, browser, PATHBUF);
+
+	if (!browser) browser = HANDBOOK_BROWSER;
+
+	argv[0] = browser;
+	i = spawn_process(argv, NULL);
+#endif
+	if (i) alert_box( _("Error"),
+		_("There was a problem running the HTML browser.  You need to set the correct program name in the Preferences window."), NULL);
+	return (i);
+}
