@@ -1,5 +1,5 @@
 /*	spawn.h
-	Copyright (C) 2007-2008 Mark Tyler and Dmitry Groshev
+	Copyright (C) 2007-2010 Mark Tyler and Dmitry Groshev
 
 	This file is part of mtPaint.
 
@@ -36,9 +36,11 @@ void spawn_quit();	// Delete temp files
 
 // Default action codes
 #define DA_GIF_CREATE  0
-#define DA_GIF_EXPLODE 1
+//#define DA_GIF_EXPLODE 1
 #define DA_GIF_PLAY    2
 #define DA_GIF_EDIT    3
 
 //	Run some default action
 int run_def_action(int action, char *sname, char *dname, int delay);
+
+int show_html(char *browser, char *docs);
