@@ -1,5 +1,5 @@
 /*	viewer.c
-	Copyright (C) 2004-2009 Mark Tyler and Dmitry Groshev
+	Copyright (C) 2004-2011 Mark Tyler and Dmitry Groshev
 
 	This file is part of mtPaint.
 
@@ -51,20 +51,14 @@ static gint viewer_keypress( GtkWidget *widget, GdkEventKey *event )
 void cline_select(GtkWidget *clist, gint row, gint col, GdkEvent *event, gpointer *pointer)
 {
 	static int last_row; // 0 initially
-	int i = row + file_arg_start, change = 0;
 
 	if (row != last_row)
 	{
-		if ( layers_total==0 )
-			change = check_for_changes();
-		else
-			change = check_layers_for_changes();
-		if ((change == 2) || (change == -10))	// Load requested file
-		{
-			do_a_load(global_argv[i]);
-			last_row = row;
-		}
-		else clist_reselect_row(GTK_CLIST(clist), last_row); // Go back
+		if ((layers_total ? check_layers_for_changes() :
+			check_for_changes()) == 1)
+			clist_reselect_row(GTK_CLIST(clist), last_row); // Go back
+		// Load requested file
+		else do_a_load(file_args[last_row = row], undo_load);
 	}
 }
 
@@ -85,9 +79,9 @@ void create_cline_area( GtkWidget *vbox1 )
 	gtk_clist_set_selection_mode(GTK_CLIST(col_list), GTK_SELECTION_BROWSE);
 
 	item[0] = txt2;
-	for ( i=file_arg_start; i<(file_arg_start + files_passed); i++ )
+	for (i = 0; i < files_passed; i++)
 	{
-		gtkuncpy(txt2, global_argv[i], PATHTXT);
+		gtkuncpy(txt2, file_args[i], PATHTXT);
 		gtk_clist_set_selectable ( GTK_CLIST(col_list),
 			gtk_clist_append(GTK_CLIST (col_list), item), TRUE );
 	}
@@ -108,7 +102,7 @@ void create_cline_area( GtkWidget *vbox1 )
 
 #include "help.c"
 
-gboolean click_help_end( GtkWidget *widget, GdkEvent *event, gpointer data )
+static gboolean click_help_end(GtkWidget *widget, GdkEvent *event, gpointer data)
 {
 	// Make sure the user can only open 1 help window
 	gtk_widget_set_sensitive(menu_widgets[MENU_HELP], TRUE);
@@ -135,13 +129,11 @@ void pressed_help()
 
 	gtk_container_set_border_width (GTK_CONTAINER (help_window), 4);
 	gtk_window_set_position (GTK_WINDOW (help_window), GTK_WIN_POS_CENTER);
-	snprintf(txt, 120, "%s - %s", VERSION, _("About"));
+	snprintf(txt, 120, "%s - %s", MT_VERSION, _("About"));
 	gtk_window_set_title (GTK_WINDOW (help_window), txt);
 	gtk_widget_set_usize (help_window, -2, 400);
 
-	box1 = gtk_vbox_new (FALSE, 0);
-	gtk_container_add (GTK_CONTAINER (help_window), box1);
-	gtk_widget_show (box1);
+	box1 = add_vbox(help_window);
 
 	box2 = xpack(box1, gtk_vbox_new(FALSE, 10));
 	gtk_container_set_border_width (GTK_CONTAINER (box2), 1);
@@ -322,8 +314,8 @@ void pan_thumbnail()		// Create thumbnail and selection box
 {
 	GtkAdjustment *hori, *vert;
 
-	while (gtk_events_pending()) gtk_main_iteration();
-		// Update main window first to get new scroll positions if necessary
+	// Update main window first to get new scroll positions if necessary
+	handle_events();
 
 	hori = gtk_scrolled_window_get_hadjustment( GTK_SCROLLED_WINDOW(scrolledwindow_canvas) );
 	vert = gtk_scrolled_window_get_vadjustment( GTK_SCROLLED_WINDOW(scrolledwindow_canvas) );
@@ -361,7 +353,7 @@ static void do_pan(GtkAdjustment *hori, GtkAdjustment *vert, int nv_h, int nv_v)
 		gtk_adjustment_value_changed(vert);
 
 		/* Process events */
-		while (gtk_events_pending()) gtk_main_iteration();
+		handle_events();
 		if (in_pan < 2) break;
 
 		/* Do delayed update */
@@ -633,8 +625,7 @@ void render_layers(unsigned char *rgb, int step, int px, int py, int pw, int ph,
 		if (ll)
 		{
 			opac = (layer_table[ll].opacity * 255 + 50) / 100;
-			xpm = ll == layer_selected ? mem_xpm_trans :
-				layer_table[ll].image->state_.xpm_trans;
+			xpm = image->trans;
 		}
 		img = image->img;
 		setup_row(xof + mx, mw, czoom, ii - i, xpm, opac,
@@ -673,8 +664,8 @@ static guint idle_focus;
 
 void vw_focus_view()						// Focus view window to main window
 {
-	int w, h, w0, h0;
-	float main_h = 0.5, main_v = 0.5, px, py, nv_h, nv_v;
+	int w0, h0;
+	float px, py, nv_h, nv_v, main_hv[2];
 	GtkAdjustment *hori, *vert;
 
 	if (idle_focus) gtk_idle_remove(idle_focus);
@@ -688,48 +679,39 @@ void vw_focus_view()						// Focus view window to main window
 		return;
 	}
 
-	hori = gtk_scrolled_window_get_hadjustment( GTK_SCROLLED_WINDOW(scrolledwindow_canvas) );
-	vert = gtk_scrolled_window_get_vadjustment( GTK_SCROLLED_WINDOW(scrolledwindow_canvas) );
-
-	canvas_size(&w, &h);
-	if (hori->page_size < w)
-		main_h = (hori->value + hori->page_size * 0.5) / w;
-	if (vert->page_size < h)
-		main_v = (vert->value + vert->page_size * 0.5) / h;
+	canvas_center(main_hv);
 
 	/* If we are editing a layer above the background make adjustments */
 	if (layers_total && layer_selected)
 	{
 		w0 = layer_table[0].image->image_.width;
 		h0 = layer_table[0].image->image_.height;
-		px = main_h * mem_width + layer_table[layer_selected].x;
-		py = main_v * mem_height + layer_table[layer_selected].y;
+		px = main_hv[0] * mem_width + layer_table[layer_selected].x;
+		py = main_hv[1] * mem_height + layer_table[layer_selected].y;
 		px = px < 0.0 ? 0.0 : px >= w0 ? w0 - 1 : px;
 		py = py < 0.0 ? 0.0 : py >= h0 ? h0 - 1 : py;
-		main_h = px / w0;
-		main_v = py / h0;
+		main_hv[0] = px / w0;
+		main_hv[1] = py / h0;
 	}
 
 	hori = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(vw_scrolledwindow));
 	vert = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(vw_scrolledwindow));
 
+	nv_h = nv_v = 0.0;
 	if (hori->page_size < vw_width)
 	{
-		nv_h = vw_width * main_h - hori->page_size * 0.5;
+		nv_h = vw_width * main_hv[0] - hori->page_size * 0.5;
 		if (nv_h + hori->page_size > vw_width)
 			nv_h = vw_width - hori->page_size;
 		if (nv_h < 0.0) nv_h = 0.0;
 	}
-	else nv_h = 0.0;
-
 	if ( vert->page_size < vw_height )
 	{
-		nv_v = vw_height * main_v - vert->page_size * 0.5;
+		nv_v = vw_height * main_hv[1] - vert->page_size * 0.5;
 		if (nv_v + vert->page_size > vw_height)
 			nv_v = vw_height - vert->page_size;
 		if (nv_v < 0.0) nv_v = 0.0;
 	}
-	else nv_v = 0.0;
 
 	/* Do nothing if nothing changed */
 	if ((hori->value == nv_h) && (vert->value == nv_v)) return;
@@ -747,7 +729,7 @@ void vw_focus_idle()
 	if (idle_focus) return;
 	if (!view_showing) return;
 	if (!vw_focus_on) return;
-	idle_focus = gtk_idle_add_priority(GTK_PRIORITY_REDRAW + 5,
+	idle_focus = threads_idle_add_priority(GTK_PRIORITY_REDRAW + 5,
 		(GtkFunction)vw_focus_view, NULL);
 }
 
@@ -771,14 +753,13 @@ gboolean vw_configure( GtkWidget *widget, GdkEventConfigure *event )
 		/* Force redraw of whole canvas as the margin has shifted */
 		gtk_widget_queue_draw(vw_drawing);
 	}
+	if (idle_focus) vw_focus_view(); // Time to refocus is NOW
 
 	return TRUE;
 }
 
-void vw_align_size( float new_zoom )
+void vw_align_size(float new_zoom)
 {
-	int sw = mem_width, sh = mem_height, i;
-
 	if (!view_showing) return;
 
 	if (new_zoom < MIN_ZOOM) new_zoom = MIN_ZOOM;
@@ -786,6 +767,15 @@ void vw_align_size( float new_zoom )
 	if (new_zoom == vw_zoom) return;
 
 	vw_zoom = new_zoom;
+	vw_realign();
+	toolbar_zoom_update();	// View zoom doesn't get changed elsewhere
+}
+
+void vw_realign()
+{
+	int sw = mem_width, sh = mem_height, i;
+
+	if (!view_showing) return;
 
 	if (layers_total && layer_selected)
 	{
@@ -809,27 +799,19 @@ void vw_align_size( float new_zoom )
 	{
 		vw_width = sw;
 		vw_height = sh;
-		gtk_widget_set_usize(vw_drawing, vw_width, vw_height);
+		wjcanvas_size(vw_drawing, vw_width, vw_height);
 	}
-	vw_focus_view();
-	toolbar_zoom_update();
+	/* !!! Let refocus wait a bit - if window is being resized, view pane's
+	 * allocation could be not yet updated (canvas is done first) - WJ */
+	vw_focus_idle();
 }
 
-void vw_realign()
-{
-	float old_zoom = vw_zoom;
-	vw_zoom = -1; /* Force resize */
-	vw_align_size(old_zoom);	// Update the view window as needed
-	vw_zoom = old_zoom; // !!! The call may have aborted early
-}
-
-void vw_repaint(int px, int py, int pw, int ph)
+static void vw_repaint(int px, int py, int pw, int ph)
 {
 	unsigned char *rgb;
+	int vport[4];
 
 	if ((pw <= 0) || (ph <= 0)) return;
-	if (px < 0) px = 0;
-	if (py < 0) py = 0;
 
 	rgb = calloc(1, pw * ph * 3);
 	if (rgb)
@@ -837,29 +819,29 @@ void vw_repaint(int px, int py, int pw, int ph)
 		memset(rgb, mem_background, pw * ph * 3);
 		view_render_rgb(rgb, px - margin_view_x, py - margin_view_y,
 			pw, ph, vw_zoom);
-		gdk_draw_rgb_image (vw_drawing->window, vw_drawing->style->black_gc,
-			px, py, pw, ph, GDK_RGB_DITHER_NONE, rgb, pw * 3);
+		wjcanvas_get_vport(vw_drawing, vport);
+		gdk_draw_rgb_image(vw_drawing->window, vw_drawing->style->black_gc,
+			px - vport[0], py - vport[1], pw, ph,
+			GDK_RGB_DITHER_NONE, rgb, pw * 3);
 		free(rgb);
 	}
 }
 
-static gint vw_expose( GtkWidget *widget, GdkEventExpose *event )
+#define REPAINT_VIEW_COST 256
+
+static gboolean vw_expose(GtkWidget *widget, GdkEventExpose *event)
 {
-	int px, py, pw, ph;
+	int vport[4];
 
-	px = event->area.x;
-	py = event->area.y;
-	pw = event->area.width;
-	ph = event->area.height;
+	wjcanvas_get_vport(widget, vport);
+	repaint_expose(event, vport, vw_repaint, REPAINT_VIEW_COST);
 
-	vw_repaint( px, py, pw, ph );
-
-	return FALSE;
+	return (FALSE);
 }
 
-void vw_update_area( int x, int y, int w, int h )	// Update x,y,w,h area of current image
+void vw_update_area(int x, int y, int w, int h)	// Update x,y,w,h area of current image
 {
-	int zoom, scale;
+	int zoom, scale, vport[4], rxy[4];
 
 	if (!view_showing) return;
 	
@@ -889,8 +871,12 @@ void vw_update_area( int x, int y, int w, int h )	// Update x,y,w,h area of curr
 		h *= scale;
 	}
 
-	gtk_widget_queue_draw_area(vw_drawing,
-		x + margin_view_x, y + margin_view_y, w, h);
+	x += margin_view_x; y += margin_view_y;
+	wjcanvas_get_vport(vw_drawing, vport);
+	if (clip(rxy, x, y, x + w, y + h, vport))
+		gtk_widget_queue_draw_area(vw_drawing,
+			rxy[0] - vport[0], rxy[1] - vport[1],
+			rxy[2] - rxy[0], rxy[3] - rxy[1]);
 }
 
 static void vw_mouse_event(int event, int x, int y, guint state, guint button)
@@ -952,8 +938,7 @@ static void vw_mouse_event(int event, int x, int y, guint state, guint button)
 					if (img[CHN_ALPHA][ofs] < (bpp == 1 ? 255 : 1))
 						continue;
 				}
-				tpix = i == layer_selected ? mem_xpm_trans :
-					layer_table[i].image->state_.xpm_trans;
+				tpix = image->trans;
 				if (tpix >= 0)
 				{
 					if (bpp == 1) ppix = rgb[ofs];
@@ -972,11 +957,11 @@ static void vw_mouse_event(int event, int x, int y, guint state, guint button)
 	}
 }
 
-static gint view_window_motion( GtkWidget *widget, GdkEventMotion *event )
+static gboolean view_window_motion(GtkWidget *widget, GdkEventMotion *event)
 {
-	int x, y;
 	GdkModifierType state;
 	guint button = 0;
+	int x, y, vport[4];
 
 	if (event->is_hint) gdk_window_get_pointer (event->window, &x, &y, &state);
 	else
@@ -992,14 +977,18 @@ static gint view_window_motion( GtkWidget *widget, GdkEventMotion *event )
 	else if (state & GDK_BUTTON3_MASK) button = 3;
 	else if (state & GDK_BUTTON2_MASK) button = 2;
 
-	vw_mouse_event(event->type, x, y, state, button);
+	/* If cursor got warped, will have another movement event to handle */
+	if (button && wjcanvas_bind_mouse(widget, event, x, y)) return (TRUE);
 
-	return TRUE;
+	wjcanvas_get_vport(widget, vport);
+	vw_mouse_event(event->type, x + vport[0], y + vport[1], state, button);
+
+	return (TRUE);
 }
 
 static gint view_window_button( GtkWidget *widget, GdkEventButton *event )
 {
-	int pflag = event->type != GDK_BUTTON_RELEASE;
+	int vport[4], pflag = event->type != GDK_BUTTON_RELEASE;
 
 	/* Steal focus from dock window */
 	if (pflag && dock_focused())
@@ -1008,7 +997,9 @@ static gint view_window_button( GtkWidget *widget, GdkEventButton *event )
 		return (TRUE);
 	}
 
-	vw_mouse_event(event->type, event->x, event->y, event->state, event->button);
+	wjcanvas_get_vport(widget, vport);
+	vw_mouse_event(event->type, event->x + vport[0], event->y + vport[1],
+		event->state, event->button);
 
 	return (pflag);
 }
@@ -1265,7 +1256,7 @@ void render_text( GtkWidget *widget )
 	PangoLayout *layout;
 	PangoFontDescription *font_desc;
 	int tx = PAD_SIZE, ty = PAD_SIZE;
-#if GTK_MINOR_VERSION >= 6
+#if GTK2VERSION >= 6 /* GTK+ 2.6+ */
 	PangoMatrix matrix = PANGO_MATRIX_INIT;
 	double degs, angle;
 	int w2, h2;
@@ -1281,7 +1272,7 @@ void render_text( GtkWidget *widget )
 
 	pango_layout_set_text( layout, inifile_get( "textString", "" ), -1 );
 
-#if GTK_MINOR_VERSION >= 6
+#if GTK2VERSION >= 6 /* GTK+ 2.6+ */
 	if (rotate)		// Rotation Toggle
 	{
 		degs = inifile_get_gint32("fontAngle", 0) * 0.01;
@@ -1341,17 +1332,10 @@ void render_text( GtkWidget *widget )
 	else have_rgb = make_text_clipboard(buf, width, height, 3);
 
 	if (have_rgb) text_paste = TEXT_PASTE_GTK;
-	else alert_box( _("Error"), _("Not enough memory to create clipboard"),
-		_("OK"), NULL, NULL );
+	else alert_box(_("Error"), _("Not enough memory to create clipboard"), NULL);
 }
 
-static gint delete_text( GtkWidget *widget, GdkEvent *event, gpointer data )
-{
-	gtk_widget_destroy( text_window );
-	return FALSE;
-}
-
-static gint paste_text_ok( GtkWidget *widget, GdkEvent *event, gpointer data )
+static void paste_text_ok(GtkWidget *widget)
 {
 	gboolean antialias[3] = { gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(text_toggle[0])),
 		gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(text_toggle[1])),
@@ -1362,11 +1346,11 @@ static gint paste_text_ok( GtkWidget *widget, GdkEvent *event, gpointer data )
 		*t_font_name = gtk_font_selection_get_font_name( GTK_FONT_SELECTION(text_font_window) );
 
 #if GTK_MAJOR_VERSION == 1
-	if ( gtk_font_selection_get_font( GTK_FONT_SELECTION(text_font_window) ) == NULL )
-		return FALSE;
+	if (!gtk_font_selection_get_font(GTK_FONT_SELECTION(text_font_window)))
+		return;
 #endif
 
-#if GTK_CHECK_VERSION(2,6,0)
+#if GTK2VERSION >= 6 /* GTK+ 2.6+ */
 	inifile_set_gint32("fontAngle", rint(read_float_spin(text_spin[1]) * 100.0));
 #endif
 
@@ -1385,11 +1369,9 @@ static gint paste_text_ok( GtkWidget *widget, GdkEvent *event, gpointer data )
 	update_stuff(UPD_XCOPY);
 	if (mem_clipboard) pressed_paste(TRUE);
 
-	delete_text( widget, event, data );
+	gtk_widget_destroy(widget);
 
 	if (t_font_name) g_free(t_font_name);
-
-	return FALSE;
 }
 
 void pressed_text()
@@ -1398,15 +1380,14 @@ void pressed_text()
 
 	text_window = add_a_window( GTK_WINDOW_TOPLEVEL, _("Paste Text"), GTK_WIN_POS_CENTER, TRUE );
 	gtk_window_set_default_size( GTK_WINDOW(text_window), 400, 400 );
-
-	vbox = gtk_vbox_new (FALSE, 0);
-	gtk_widget_show (vbox);
-	gtk_container_add (GTK_CONTAINER (text_window), vbox);
+	vbox = add_vbox(text_window);
 
 	text_font_window = xpack(vbox, gtk_font_selection_new());
 	gtk_widget_show(text_font_window);
 	gtk_container_set_border_width (GTK_CONTAINER (text_font_window), 4);
 
+	accept_ctrl_enter(GTK_FONT_SELECTION(text_font_window)->preview_entry);
+
 	add_hseparator( vbox, 200, 10 );
 
 	hbox = pack5(vbox, gtk_hbox_new(FALSE, 0));
@@ -1437,7 +1418,7 @@ void pressed_text()
 
 	text_toggle[2] = add_a_toggle( _("Angle of rotation ="), hbox, FALSE );
 
-#if GTK_CHECK_VERSION(2,6,0)
+#if GTK2VERSION >= 6 /* GTK+ 2.6+ */
 	text_spin[1] = pack5(hbox, add_float_spin(
 		inifile_get_gint32("fontAngle", 0) * 0.01, -360, 360));
 	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(text_toggle[2]), 
@@ -1450,15 +1431,15 @@ void pressed_text()
 
 	hbox = pack5(vbox, OK_box(0, text_window,
 		_("Paste Text"), GTK_SIGNAL_FUNC(paste_text_ok),
-		_("Cancel"), GTK_SIGNAL_FUNC(delete_text)));
-
-	gtk_widget_show(text_window);
-	gtk_window_set_transient_for( GTK_WINDOW(text_window), GTK_WINDOW(main_window) );
+		_("Cancel"), GTK_SIGNAL_FUNC(gtk_widget_destroy)));
 
 	gtk_font_selection_set_font_name( GTK_FONT_SELECTION(text_font_window),
 		inifile_get( "lastTextFont", "-misc-fixed-bold-r-normal-*-*-120-*-*-c-*-iso8859-1" ) );
 	gtk_font_selection_set_preview_text( GTK_FONT_SELECTION(text_font_window),
 		inifile_get( "textString", _("Enter Text Here") ) );
 
+	gtk_window_set_transient_for( GTK_WINDOW(text_window), GTK_WINDOW(main_window) );
+	gtk_widget_show(text_window);
+
 	gtk_widget_grab_focus( GTK_FONT_SELECTION(text_font_window)->preview_entry );
 }
